"""Test for `maas.client.viscera.interfaces`."""

import random

from testtools.matchers import (
    Equals,
    IsInstance,
    MatchesAll,
    MatchesSetwise,
    MatchesStructure,
)

from ..fabrics import Fabric, Fabrics
from ..interfaces import (
    Interface,
    Interfaces,
    InterfaceDiscoveredLink,
    InterfaceDiscoveredLinks,
    InterfaceLink,
    InterfaceLinks,
)
from ..nodes import Node, Nodes
from ..subnets import Subnet, Subnets
from ..vlans import Vlan, Vlans

from ..testing import bind
from ...enum import InterfaceType, LinkMode
from ...testing import make_string_without_spaces, TestCase


def make_origin():
    """
    Create a new origin with required objects.
    """
    return bind(
        Fabrics,
        Fabric,
        Interfaces,
        Interface,
        InterfaceLinks,
        InterfaceLink,
        InterfaceDiscoveredLinks,
        InterfaceDiscoveredLink,
        Nodes,
        Node,
        Subnets,
        Subnet,
        Vlans,
        Vlan,
    )


class TestInterfaces(TestCase):
    def test__read_bad_node_type(self):
        Interfaces = make_origin().Interfaces
        error = self.assertRaises(TypeError, Interfaces.read, random.randint(0, 100))
        self.assertEquals("node must be a Node or str, not int", str(error))

    def test__read_with_system_id(self):
        Interfaces = make_origin().Interfaces
        system_id = make_string_without_spaces()
        interfaces = [
            {
                "system_id": system_id,
                "id": random.randint(0, 100),
                "name": make_string_without_spaces(),
                "type": InterfaceType.PHYSICAL.value,
            }
            for _ in range(3)
        ]
        Interfaces._handler.read.return_value = interfaces
        interfaces = Interfaces.read(system_id)
        self.assertThat(len(interfaces), Equals(3))
        Interfaces._handler.read.assert_called_once_with(system_id=system_id)

    def test__read_with_Node(self):
        origin = make_origin()
        Interfaces, Node = origin.Interfaces, origin.Node
        system_id = make_string_without_spaces()
        node = Node(system_id)
        interfaces = [
            {
                "system_id": system_id,
                "id": random.randint(0, 100),
                "name": make_string_without_spaces(),
                "type": InterfaceType.PHYSICAL.value,
            }
            for _ in range(3)
        ]
        Interfaces._handler.read.return_value = interfaces
        interfaces = Interfaces.read(node)
        self.assertThat(len(interfaces), Equals(3))
        Interfaces._handler.read.assert_called_once_with(system_id=system_id)

    def test__create_bad_node_type(self):
        Interfaces = make_origin().Interfaces
        error = self.assertRaises(TypeError, Interfaces.create, random.randint(0, 100))
        self.assertEquals("node must be a Node or str, not int", str(error))

    def test__create_bad_vlan_type(self):
        Interfaces = make_origin().Interfaces
        error = self.assertRaises(
            TypeError,
            Interfaces.create,
            make_string_without_spaces(),
            vlan=make_string_without_spaces(),
        )
        self.assertEquals("vlan must be a Vlan or int, not str", str(error))

    def test__create_bad_interface_type_type(self):
        Interfaces = make_origin().Interfaces
        error = self.assertRaises(
            TypeError,
            Interfaces.create,
            make_string_without_spaces(),
            interface_type=make_string_without_spaces(),
        )
        self.assertEquals(
            "interface_type must be an InterfaceType, not str", str(error)
        )

    def test__create_physical_requires_mac_address(self):
        Interfaces = make_origin().Interfaces
        error = self.assertRaises(
            ValueError, Interfaces.create, make_string_without_spaces()
        )
        self.assertEquals("mac_address required for physical interface", str(error))

    def test__create_physical_with_all_values(self):
        origin = make_origin()
        Interfaces, Interface = origin.Interfaces, origin.Interface
        system_id = make_string_without_spaces()
        mac_address = "00:11:22:33:44:55"
        name = make_string_without_spaces()
        tags = [make_string_without_spaces() for _ in range(3)]
        mtu = random.randint(1500, 3000)
        vlan = random.randint(1, 20)
        accept_ra = random.choice([True, False])
        autoconf = random.choice([True, False])
        interface_data = {
            "system_id": system_id,
            "id": random.randint(1, 20),
            "type": InterfaceType.PHYSICAL.value,
            "name": name,
            "tags": tags,
        }
        Interfaces._handler.create_physical.return_value = interface_data
        nic = Interfaces.create(
            node=system_id,
            mac_address=mac_address,
            name=name,
            tags=tags,
            mtu=mtu,
            vlan=vlan,
            accept_ra=accept_ra,
            autoconf=autoconf,
        )
        self.assertThat(nic, IsInstance(Interface))
        Interfaces._handler.create_physical.assert_called_once_with(
            system_id=system_id,
            mac_address=mac_address,
            name=name,
            tags=tags,
            mtu=mtu,
            vlan=vlan,
            accept_ra=accept_ra,
            autoconf=autoconf,
        )

    def test__create_physical_with_objects(self):
        origin = make_origin()
        Interfaces, Interface = origin.Interfaces, origin.Interface
        system_id = make_string_without_spaces()
        node = origin.Node(system_id)
        mac_address = "00:11:22:33:44:55"
        name = make_string_without_spaces()
        tags = [make_string_without_spaces() for _ in range(3)]
        mtu = random.randint(1500, 3000)
        vlan_id = random.randint(1, 20)
        vlan = origin.Vlan({"fabric_id": random.randint(1, 20), "id": vlan_id})
        accept_ra = random.choice([True, False])
        autoconf = random.choice([True, False])
        interface_data = {
            "system_id": system_id,
            "id": random.randint(1, 20),
            "type": InterfaceType.PHYSICAL.value,
            "name": name,
            "tags": tags,
        }
        Interfaces._handler.create_physical.return_value = interface_data
        nic = Interfaces.create(
            node=node,
            mac_address=mac_address,
            name=name,
            tags=tags,
            mtu=mtu,
            vlan=vlan,
            accept_ra=accept_ra,
            autoconf=autoconf,
        )
        self.assertThat(nic, IsInstance(Interface))
        Interfaces._handler.create_physical.assert_called_once_with(
            system_id=system_id,
            mac_address=mac_address,
            name=name,
            tags=tags,
            mtu=mtu,
            vlan=vlan_id,
            accept_ra=accept_ra,
            autoconf=autoconf,
        )

    def test__create_bond_fails_with_parent(self):
        Interfaces = make_origin().Interfaces
        error = self.assertRaises(
            ValueError,
            Interfaces.create,
            make_string_without_spaces(),
            InterfaceType.BOND,
            name=make_string_without_spaces(),
            parent=make_string_without_spaces(),
        )
        self.assertEquals("use parents not parent for bond interface", str(error))

    def test__create_bond_fails_parents_not_iterable(self):
        Interfaces = make_origin().Interfaces
        error = self.assertRaises(
            TypeError,
            Interfaces.create,
            make_string_without_spaces(),
            InterfaceType.BOND,
            name=make_string_without_spaces(),
            parents=random.randint(1, 20),
        )
        self.assertEquals("parents must be a iterable, not int", str(error))

    def test__create_bond_fails_parents_is_not_Interface_or_int(self):
        Interfaces = make_origin().Interfaces
        error = self.assertRaises(
            TypeError,
            Interfaces.create,
            make_string_without_spaces(),
            InterfaceType.BOND,
            name=make_string_without_spaces(),
            parents=[make_string_without_spaces()],
        )
        self.assertEquals("parent[0] must be an Interface or int, not str", str(error))

    def test__create_bond_fails_name_missing(self):
        Interfaces = make_origin().Interfaces
        error = self.assertRaises(
            ValueError,
            Interfaces.create,
            make_string_without_spaces(),
            InterfaceType.BOND,
            parents=[random.randint(1, 20)],
        )
        self.assertEquals("name is required for bond interface", str(error))

    def test__create_bond_with_all_values(self):
        origin = make_origin()
        Interfaces, Interface = origin.Interfaces, origin.Interface
        system_id = make_string_without_spaces()
        mac_address = "00:11:22:33:44:55"
        name = make_string_without_spaces()
        tags = [make_string_without_spaces() for _ in range(3)]
        mtu = random.randint(1500, 3000)
        vlan = random.randint(1, 20)
        accept_ra = random.choice([True, False])
        autoconf = random.choice([True, False])
        bond_mode = make_string_without_spaces()
        bond_miimon = random.randint(1, 100)
        bond_downdelay = random.randint(1, 10)
        bond_updelay = random.randint(1, 10)
        bond_lacp_rate = random.choice(["fast", "slow"])
        bond_xmit_hash_policy = make_string_without_spaces()
        parents = [random.randint(1, 10) for _ in range(3)]
        interface_data = {
            "system_id": system_id,
            "id": random.randint(1, 20),
            "type": InterfaceType.BOND.value,
            "name": name,
            "tags": tags,
        }
        Interfaces._handler.create_bond.return_value = interface_data
        nic = Interfaces.create(
            node=system_id,
            interface_type=InterfaceType.BOND,
            parents=parents,
            mac_address=mac_address,
            name=name,
            tags=tags,
            mtu=mtu,
            vlan=vlan,
            accept_ra=accept_ra,
            autoconf=autoconf,
            bond_mode=bond_mode,
            bond_miimon=bond_miimon,
            bond_downdelay=bond_downdelay,
            bond_updelay=bond_updelay,
            bond_lacp_rate=bond_lacp_rate,
            bond_xmit_hash_policy=bond_xmit_hash_policy,
        )
        self.assertThat(nic, IsInstance(Interface))
        Interfaces._handler.create_bond.assert_called_once_with(
            system_id=system_id,
            parents=parents,
            mac_address=mac_address,
            name=name,
            tags=tags,
            mtu=mtu,
            vlan=vlan,
            accept_ra=accept_ra,
            autoconf=autoconf,
            bond_mode=bond_mode,
            bond_miimon=bond_miimon,
            bond_downdelay=bond_downdelay,
            bond_updelay=bond_updelay,
            bond_lacp_rate=bond_lacp_rate,
            bond_xmit_hash_policy=bond_xmit_hash_policy,
        )

    def test__create_bond_with_objects(self):
        origin = make_origin()
        Interfaces, Interface = origin.Interfaces, origin.Interface
        system_id = make_string_without_spaces()
        node = origin.Node(system_id)
        mac_address = "00:11:22:33:44:55"
        name = make_string_without_spaces()
        tags = [make_string_without_spaces() for _ in range(3)]
        mtu = random.randint(1500, 3000)
        vlan_id = random.randint(1, 10)
        vlan = origin.Vlan({"fabric_id": random.randint(1, 20), "id": vlan_id})
        accept_ra = random.choice([True, False])
        autoconf = random.choice([True, False])
        bond_mode = make_string_without_spaces()
        bond_miimon = random.randint(1, 100)
        bond_downdelay = random.randint(1, 10)
        bond_updelay = random.randint(1, 10)
        bond_lacp_rate = random.choice(["fast", "slow"])
        bond_xmit_hash_policy = make_string_without_spaces()
        parents = [random.randint(1, 10) for _ in range(3)]
        parent_objs = [Interface((system_id, parent)) for parent in parents]
        interface_data = {
            "system_id": system_id,
            "id": random.randint(1, 20),
            "type": InterfaceType.BOND.value,
            "name": name,
            "tags": tags,
        }
        Interfaces._handler.create_bond.return_value = interface_data
        nic = Interfaces.create(
            node=node,
            interface_type=InterfaceType.BOND,
            parents=parent_objs,
            mac_address=mac_address,
            name=name,
            tags=tags,
            mtu=mtu,
            vlan=vlan,
            accept_ra=accept_ra,
            autoconf=autoconf,
            bond_mode=bond_mode,
            bond_miimon=bond_miimon,
            bond_downdelay=bond_downdelay,
            bond_updelay=bond_updelay,
            bond_lacp_rate=bond_lacp_rate,
            bond_xmit_hash_policy=bond_xmit_hash_policy,
        )
        self.assertThat(nic, IsInstance(Interface))
        Interfaces._handler.create_bond.assert_called_once_with(
            system_id=system_id,
            parents=parents,
            mac_address=mac_address,
            name=name,
            tags=tags,
            mtu=mtu,
            vlan=vlan_id,
            accept_ra=accept_ra,
            autoconf=autoconf,
            bond_mode=bond_mode,
            bond_miimon=bond_miimon,
            bond_downdelay=bond_downdelay,
            bond_updelay=bond_updelay,
            bond_lacp_rate=bond_lacp_rate,
            bond_xmit_hash_policy=bond_xmit_hash_policy,
        )

    def test__create_vlan_fails_with_parents(self):
        Interfaces = make_origin().Interfaces
        error = self.assertRaises(
            ValueError,
            Interfaces.create,
            make_string_without_spaces(),
            InterfaceType.VLAN,
            parents=[make_string_without_spaces()],
        )
        self.assertEquals("use parent not parents for VLAN interface", str(error))

    def test__create_vlan_fails_without_parent(self):
        Interfaces = make_origin().Interfaces
        error = self.assertRaises(
            ValueError,
            Interfaces.create,
            make_string_without_spaces(),
            InterfaceType.VLAN,
        )
        self.assertEquals("parent is required for VLAN interface", str(error))

    def test__create_vlan_fails_parent_wrong_type(self):
        Interfaces = make_origin().Interfaces
        error = self.assertRaises(
            TypeError,
            Interfaces.create,
            make_string_without_spaces(),
            InterfaceType.VLAN,
            parent=make_string_without_spaces(),
        )
        self.assertEquals("parent must be an Interface or int, not str", str(error))

    def test__create_vlan_fails_without_vlan(self):
        Interfaces = make_origin().Interfaces
        error = self.assertRaises(
            ValueError,
            Interfaces.create,
            make_string_without_spaces(),
            InterfaceType.VLAN,
            parent=random.randint(1, 10),
        )
        self.assertEquals("vlan is required for VLAN interface", str(error))

    def test__create_vlan_with_all_values(self):
        origin = make_origin()
        Interfaces, Interface = origin.Interfaces, origin.Interface
        system_id = make_string_without_spaces()
        tags = [make_string_without_spaces() for _ in range(3)]
        mtu = random.randint(1500, 3000)
        vlan = random.randint(1, 20)
        accept_ra = random.choice([True, False])
        autoconf = random.choice([True, False])
        parent = random.randint(1, 10)
        interface_data = {
            "system_id": system_id,
            "id": random.randint(1, 20),
            "type": InterfaceType.VLAN.value,
            "tags": tags,
        }
        Interfaces._handler.create_vlan.return_value = interface_data
        nic = Interfaces.create(
            node=system_id,
            interface_type=InterfaceType.VLAN,
            parent=parent,
            tags=tags,
            mtu=mtu,
            vlan=vlan,
            accept_ra=accept_ra,
            autoconf=autoconf,
        )
        self.assertThat(nic, IsInstance(Interface))
        Interfaces._handler.create_vlan.assert_called_once_with(
            system_id=system_id,
            parent=parent,
            tags=tags,
            mtu=mtu,
            vlan=vlan,
            accept_ra=accept_ra,
            autoconf=autoconf,
        )

    def test__create_vlan_with_objects(self):
        origin = make_origin()
        Interfaces, Interface = origin.Interfaces, origin.Interface
        system_id = make_string_without_spaces()
        node = origin.Node(system_id)
        tags = [make_string_without_spaces() for _ in range(3)]
        mtu = random.randint(1500, 3000)
        vlan_id = random.randint(1, 10)
        vlan = origin.Vlan({"fabric_id": random.randint(1, 20), "id": vlan_id})
        accept_ra = random.choice([True, False])
        autoconf = random.choice([True, False])
        parent_id = random.randint(1, 10)
        parent_obj = Interface((system_id, parent_id))
        interface_data = {
            "system_id": system_id,
            "id": random.randint(1, 20),
            "type": InterfaceType.VLAN.value,
            "tags": tags,
        }
        Interfaces._handler.create_vlan.return_value = interface_data
        nic = Interfaces.create(
            node=node,
            interface_type=InterfaceType.VLAN,
            parent=parent_obj,
            tags=tags,
            mtu=mtu,
            vlan=vlan,
            accept_ra=accept_ra,
            autoconf=autoconf,
        )
        self.assertThat(nic, IsInstance(Interface))
        Interfaces._handler.create_vlan.assert_called_once_with(
            system_id=system_id,
            parent=parent_id,
            tags=tags,
            mtu=mtu,
            vlan=vlan_id,
            accept_ra=accept_ra,
            autoconf=autoconf,
        )

    def test__create_bridge_fails_with_parents(self):
        Interfaces = make_origin().Interfaces
        error = self.assertRaises(
            ValueError,
            Interfaces.create,
            make_string_without_spaces(),
            InterfaceType.BRIDGE,
            parents=[make_string_without_spaces()],
        )
        self.assertEquals("use parent not parents for bridge interface", str(error))

    def test__create_bridge_fails_without_parent(self):
        Interfaces = make_origin().Interfaces
        error = self.assertRaises(
            ValueError,
            Interfaces.create,
            make_string_without_spaces(),
            InterfaceType.BRIDGE,
        )
        self.assertEquals("parent is required for bridge interface", str(error))

    def test__create_bridge_fails_parent_wrong_type(self):
        Interfaces = make_origin().Interfaces
        error = self.assertRaises(
            TypeError,
            Interfaces.create,
            make_string_without_spaces(),
            InterfaceType.BRIDGE,
            parent=make_string_without_spaces(),
        )
        self.assertEquals("parent must be an Interface or int, not str", str(error))

    def test__create_bridge_fails_missing_name(self):
        Interfaces = make_origin().Interfaces
        error = self.assertRaises(
            ValueError,
            Interfaces.create,
            make_string_without_spaces(),
            InterfaceType.BRIDGE,
            parent=random.randint(1, 10),
        )
        self.assertEquals("name is required for bridge interface", str(error))

    def test__create_bridge_with_all_values(self):
        origin = make_origin()
        Interfaces, Interface = origin.Interfaces, origin.Interface
        system_id = make_string_without_spaces()
        name = make_string_without_spaces()
        mac_address = "00:11:22:33:44:55"
        tags = [make_string_without_spaces() for _ in range(3)]
        mtu = random.randint(1500, 3000)
        vlan = random.randint(1, 20)
        accept_ra = random.choice([True, False])
        autoconf = random.choice([True, False])
        parent = random.randint(1, 10)
        bridge_stp = random.choice([True, False])
        bridge_fd = random.randint(1, 10)
        interface_data = {
            "system_id": system_id,
            "id": random.randint(1, 20),
            "type": InterfaceType.BRIDGE.value,
            "name": name,
            "tags": tags,
        }
        Interfaces._handler.create_bridge.return_value = interface_data
        nic = Interfaces.create(
            node=system_id,
            interface_type=InterfaceType.BRIDGE,
            parent=parent,
            name=name,
            tags=tags,
            mtu=mtu,
            vlan=vlan,
            accept_ra=accept_ra,
            autoconf=autoconf,
            mac_address=mac_address,
            bridge_stp=bridge_stp,
            bridge_fd=bridge_fd,
        )
        self.assertThat(nic, IsInstance(Interface))
        Interfaces._handler.create_bridge.assert_called_once_with(
            system_id=system_id,
            parent=parent,
            name=name,
            tags=tags,
            mtu=mtu,
            vlan=vlan,
            accept_ra=accept_ra,
            autoconf=autoconf,
            mac_address=mac_address,
            bridge_stp=bridge_stp,
            bridge_fd=bridge_fd,
        )

    def test__create_bridge_with_objects(self):
        origin = make_origin()
        Interfaces, Interface = origin.Interfaces, origin.Interface
        system_id = make_string_without_spaces()
        node = origin.Node(system_id)
        name = make_string_without_spaces()
        mac_address = "00:11:22:33:44:55"
        tags = [make_string_without_spaces() for _ in range(3)]
        mtu = random.randint(1500, 3000)
        vlan_id = random.randint(1, 10)
        vlan = origin.Vlan({"fabric_id": random.randint(1, 20), "id": vlan_id})
        accept_ra = random.choice([True, False])
        autoconf = random.choice([True, False])
        parent_id = random.randint(1, 10)
        parent_obj = Interface((system_id, parent_id))
        bridge_stp = random.choice([True, False])
        bridge_fd = random.randint(1, 10)
        interface_data = {
            "system_id": system_id,
            "id": random.randint(1, 20),
            "type": InterfaceType.VLAN.value,
            "name": name,
            "tags": tags,
        }
        Interfaces._handler.create_bridge.return_value = interface_data
        nic = Interfaces.create(
            node=node,
            interface_type=InterfaceType.BRIDGE,
            parent=parent_obj,
            name=name,
            tags=tags,
            mtu=mtu,
            vlan=vlan,
            accept_ra=accept_ra,
            autoconf=autoconf,
            mac_address=mac_address,
            bridge_stp=bridge_stp,
            bridge_fd=bridge_fd,
        )
        self.assertThat(nic, IsInstance(Interface))
        Interfaces._handler.create_bridge.assert_called_once_with(
            system_id=system_id,
            parent=parent_id,
            name=name,
            tags=tags,
            mtu=mtu,
            vlan=vlan_id,
            accept_ra=accept_ra,
            autoconf=autoconf,
            mac_address=mac_address,
            bridge_stp=bridge_stp,
            bridge_fd=bridge_fd,
        )

    def test__create_unknown_fails(self):
        Interfaces = make_origin().Interfaces
        error = self.assertRaises(
            ValueError,
            Interfaces.create,
            make_string_without_spaces(),
            InterfaceType.UNKNOWN,
        )
        self.assertEquals(
            "cannot create an interface of type: %s" % InterfaceType.UNKNOWN, str(error)
        )

    def test__by_name(self):
        origin = make_origin()
        Interfaces, Interface = origin.Interfaces, origin.Interface
        system_id = make_string_without_spaces()
        eth0_Interface = Interface(
            {"system_id": system_id, "id": random.randint(0, 100), "name": "eth0"}
        )
        eth1_Interface = Interface(
            {"system_id": system_id, "id": random.randint(0, 100), "name": "eth1"}
        )
        interfaces = Interfaces([eth0_Interface, eth1_Interface])
        self.assertEquals(
            {"eth0": eth0_Interface, "eth1": eth1_Interface}, interfaces.by_name
        )

    def test__get_by_name(self):
        origin = make_origin()
        Interfaces, Interface = origin.Interfaces, origin.Interface
        system_id = make_string_without_spaces()
        eth0_Interface = Interface(
            {"system_id": system_id, "id": random.randint(0, 100), "name": "eth0"}
        )
        eth1_Interface = Interface(
            {"system_id": system_id, "id": random.randint(0, 100), "name": "eth1"}
        )
        interfaces = Interfaces([eth0_Interface, eth1_Interface])
        eth0 = interfaces.get_by_name("eth0")
        self.assertEquals(eth0, eth0_Interface)


class TestInterface(TestCase):
    def test__interface_read_system_id(self):
        Interface = make_origin().Interface
        system_id = make_string_without_spaces()
        interface = {
            "system_id": system_id,
            "id": random.randint(0, 100),
            "name": make_string_without_spaces(),
            "type": InterfaceType.PHYSICAL.value,
        }
        Interface._handler.read.return_value = interface
        self.assertThat(
            Interface.read(node=system_id, id=interface["id"]),
            Equals(Interface(interface)),
        )
        Interface._handler.read.assert_called_once_with(
            system_id=system_id, id=interface["id"]
        )

    def test__interface_read_Node(self):
        origin = make_origin()
        Interface = origin.Interface
        system_id = make_string_without_spaces()
        interface = {
            "system_id": system_id,
            "id": random.randint(0, 100),
            "name": make_string_without_spaces(),
            "type": InterfaceType.PHYSICAL.value,
        }
        Interface._handler.read.return_value = interface
        self.assertThat(
            Interface.read(node=origin.Node(system_id), id=interface["id"]),
            Equals(Interface(interface)),
        )
        Interface._handler.read.assert_called_once_with(
            system_id=system_id, id=interface["id"]
        )

    def test__interface_read_TypeError(self):
        Interface = make_origin().Interface
        error = self.assertRaises(
            TypeError, Interface.read, random.randint(0, 100), random.randint(0, 100)
        )
        self.assertEquals("node must be a Node or str, not int", str(error))

    def test__interface_save_tags(self):
        Interface = make_origin().Interface
        Interface._handler.params = ["system_id", "id"]
        interface = Interface(
            {
                "system_id": make_string_without_spaces(),
                "id": random.randint(0, 100),
                "tags": ["keep", "update", "delete"],
                "params": {},
            }
        )
        del interface.tags[2]
        interface.tags[1] = "updated"
        Interface._handler.update.return_value = {
            "system_id": interface.node.system_id,
            "id": interface.id,
            "tags": ["keep", "updated"],
            "params": {},
        }
        interface.save()
        Interface._handler.update.assert_called_once_with(
            system_id=interface.node.system_id, id=interface.id, tags="keep,updated"
        )

    def test__interface_doesnt_save_tags_if_same_diff_order(self):
        Interface = make_origin().Interface
        Interface._handler.params = ["system_id", "id"]
        interface = Interface(
            {
                "system_id": make_string_without_spaces(),
                "id": random.randint(0, 100),
                "tags": ["keep", "update", "delete"],
                "params": {},
            }
        )
        interface.tags = ["delete", "keep", "update"]
        interface.save()
        self.assertEqual(0, Interface._handler.update.call_count)

    def test__interface_save_passes_parameters(self):
        Interface = make_origin().Interface
        Interface._handler.params = ["system_id", "id"]
        interface = Interface(
            {
                "system_id": make_string_without_spaces(),
                "id": random.randint(0, 100),
                "tags": [],
                "params": {"mtu": 1500},
            }
        )
        interface.params["mtu"] = 3000
        Interface._handler.update.return_value = {
            "system_id": interface.node.system_id,
            "id": interface.id,
            "tags": [],
            "params": {"mtu": 3000},
        }
        interface.save()
        Interface._handler.update.assert_called_once_with(
            system_id=interface.node.system_id, id=interface.id, mtu=3000
        )

    def test__interface_save_handles_str_params(self):
        Interface = make_origin().Interface
        Interface._handler.params = ["system_id", "id"]
        interface = Interface(
            {
                "system_id": make_string_without_spaces(),
                "id": random.randint(0, 100),
                "tags": [],
                "params": "",
            }
        )
        interface.params = {"mtu": 3000}
        Interface._handler.update.return_value = {
            "system_id": interface.node.system_id,
            "id": interface.id,
            "tags": [],
            "params": {"mtu": 3000},
        }
        interface.save()
        Interface._handler.update.assert_called_once_with(
            system_id=interface.node.system_id, id=interface.id, mtu=3000
        )

    def test__interface_save_sets_vlan_to_None(self):
        Interface = make_origin().Interface
        Interface._handler.params = ["system_id", "id"]
        vlan_data = {
            "fabric_id": random.randint(0, 100),
            "id": random.randint(0, 100),
            "vid": random.randint(0, 100),
        }
        interface = Interface(
            {
                "system_id": make_string_without_spaces(),
                "id": random.randint(0, 100),
                "tags": [],
                "params": {},
                "vlan": vlan_data,
            }
        )
        interface.vlan = None
        Interface._handler.update.return_value = {
            "system_id": interface.node.system_id,
            "id": interface.id,
            "tags": [],
            "params": {},
            "vlan": None,
        }
        interface.save()
        Interface._handler.update.assert_called_once_with(
            system_id=interface.node.system_id, id=interface.id, vlan=None
        )

    def test__interface_save_sets_vlan_to_new_vlan(self):
        origin = make_origin()
        Interface, Vlan = origin.Interface, origin.Vlan
        Interface._handler.params = ["system_id", "id"]
        vlan_data = {
            "fabric_id": random.randint(0, 100),
            "id": random.randint(0, 100),
            "vid": random.randint(0, 100),
        }
        interface = Interface(
            {
                "system_id": make_string_without_spaces(),
                "id": random.randint(0, 100),
                "tags": [],
                "params": {},
                "vlan": vlan_data,
            }
        )
        new_vlan = Vlan(
            {
                "fabric_id": random.randint(0, 100),
                "id": random.randint(101, 200),
                "vid": random.randint(0, 100),
            }
        )
        interface.vlan = new_vlan
        Interface._handler.update.return_value = {
            "system_id": interface.node.system_id,
            "id": interface.id,
            "tags": [],
            "params": {},
            "vlan": new_vlan._data,
        }
        interface.save()
        Interface._handler.update.assert_called_once_with(
            system_id=interface.node.system_id, id=interface.id, vlan=new_vlan.id
        )

    def test__interface_save_doesnt_change_vlan_when_same(self):
        origin = make_origin()
        Interface, Vlan = origin.Interface, origin.Vlan
        Interface._handler.params = ["system_id", "id"]
        vlan_data = {
            "fabric_id": random.randint(0, 100),
            "id": random.randint(0, 100),
            "vid": random.randint(0, 100),
        }
        vlan = Vlan(dict(vlan_data))
        interface = Interface(
            {
                "system_id": make_string_without_spaces(),
                "id": random.randint(0, 100),
                "tags": [],
                "params": {},
                "vlan": vlan_data,
            }
        )
        interface.vlan = vlan
        interface.save()
        self.assertEqual(0, Interface._handler.update.call_count)

    def test__interface_delete(self):
        Interface = make_origin().Interface
        system_id = make_string_without_spaces()
        interface_data = {
            "system_id": system_id,
            "id": random.randint(0, 100),
            "name": make_string_without_spaces(),
            "type": InterfaceType.PHYSICAL.value,
        }
        interface = Interface(interface_data)
        interface.delete()
        Interface._handler.delete.assert_called_once_with(
            system_id=system_id, id=interface_data["id"]
        )

    def test__interface_disconnect(self):
        Interface = make_origin().Interface
        system_id = make_string_without_spaces()
        interface_data = {
            "system_id": system_id,
            "id": random.randint(0, 100),
            "name": make_string_without_spaces(),
            "type": InterfaceType.PHYSICAL.value,
            "links": [{"id": random.randint(0, 100), "mode": LinkMode.AUTO.value}],
        }
        interface = Interface(interface_data)
        updated_data = dict(interface_data)
        updated_data["links"] = []
        Interface._handler.disconnect.return_value = updated_data
        interface.disconnect()
        Interface._handler.disconnect.assert_called_once_with(
            system_id=system_id, id=interface_data["id"]
        )
        self.assertEquals([], list(interface.links))

    def test__interface_links_create_raises_TypeError_no_Interface(self):
        origin = make_origin()
        InterfaceLinks = origin.InterfaceLinks
        error = self.assertRaises(
            TypeError, InterfaceLinks.create, random.randint(0, 1000), LinkMode.AUTO
        )
        self.assertEquals("interface must be an Interface, not int", str(error))

    def test__interface_links_create_raises_TypeError_no_LinkMode(self):
        origin = make_origin()
        Interface, InterfaceLinks = origin.Interface, origin.InterfaceLinks
        interface = Interface(
            {"system_id": make_string_without_spaces(), "id": random.randint(0, 100)}
        )
        error = self.assertRaises(
            TypeError, InterfaceLinks.create, interface, LinkMode.AUTO.value
        )
        self.assertEquals("mode must be a LinkMode, not str", str(error))

    def test__interface_links_create_raises_TypeError_no_Subnet(self):
        origin = make_origin()
        Interface, InterfaceLinks = origin.Interface, origin.InterfaceLinks
        interface = Interface(
            {"system_id": make_string_without_spaces(), "id": random.randint(0, 100)}
        )
        error = self.assertRaises(
            TypeError,
            InterfaceLinks.create,
            interface,
            LinkMode.AUTO,
            subnet=make_string_without_spaces(),
        )
        self.assertEquals("subnet must be a Subnet or int, not str", str(error))

    def test__interface_links_create_raises_ValueError_AUTO_no_Subnet(self):
        origin = make_origin()
        Interface, InterfaceLinks = origin.Interface, origin.InterfaceLinks
        interface = Interface(
            {"system_id": make_string_without_spaces(), "id": random.randint(0, 100)}
        )
        error = self.assertRaises(
            ValueError, InterfaceLinks.create, interface, LinkMode.AUTO
        )
        self.assertEquals("subnet is required for %s" % LinkMode.AUTO, str(error))

    def test__interface_links_create_raises_ValueError_STATIC_no_Subnet(self):
        origin = make_origin()
        Interface, InterfaceLinks = origin.Interface, origin.InterfaceLinks
        interface = Interface(
            {"system_id": make_string_without_spaces(), "id": random.randint(0, 100)}
        )
        error = self.assertRaises(
            ValueError, InterfaceLinks.create, interface, LinkMode.STATIC
        )
        self.assertEquals("subnet is required for %s" % LinkMode.STATIC, str(error))

    def test__interface_links_create_raises_ValueError_LINK_UP_gateway(self):
        origin = make_origin()
        Interface, InterfaceLinks = origin.Interface, origin.InterfaceLinks
        interface = Interface(
            {"system_id": make_string_without_spaces(), "id": random.randint(0, 100)}
        )
        error = self.assertRaises(
            ValueError,
            InterfaceLinks.create,
            interface,
            LinkMode.LINK_UP,
            default_gateway=True,
        )
        self.assertEquals(
            "cannot set as default_gateway for %s" % LinkMode.LINK_UP, str(error)
        )

    def test__interface_links_create_raises_ValueError_DHCP_gateway(self):
        origin = make_origin()
        Interface, InterfaceLinks = origin.Interface, origin.InterfaceLinks
        interface = Interface(
            {"system_id": make_string_without_spaces(), "id": random.randint(0, 100)}
        )
        error = self.assertRaises(
            ValueError,
            InterfaceLinks.create,
            interface,
            LinkMode.DHCP,
            default_gateway=True,
        )
        self.assertEquals(
            "cannot set as default_gateway for %s" % LinkMode.DHCP, str(error)
        )

    def test__interface_links_create_AUTO(self):
        origin = make_origin()
        Interface, Subnet = origin.Interface, origin.Subnet
        system_id = make_string_without_spaces()
        interface_data = {
            "system_id": system_id,
            "id": random.randint(0, 100),
            "name": make_string_without_spaces(),
            "type": InterfaceType.PHYSICAL.value,
            "links": [
                {
                    "id": random.randint(0, 100),
                    "mode": LinkMode.LINK_UP.value,
                    "subnet": {"id": random.randint(0, 100)},
                }
            ],
        }
        interface = Interface(interface_data)
        updated_data = dict(interface_data)
        link_id = random.randint(100, 200)
        subnet_id = random.randint(1, 100)
        updated_data["links"] = [
            {"id": link_id, "mode": LinkMode.AUTO.value, "subnet": {"id": subnet_id}}
        ]
        Interface._handler.link_subnet.return_value = updated_data
        interface.links.create(LinkMode.AUTO, subnet_id)
        Interface._handler.link_subnet.assert_called_once_with(
            system_id=interface.node.system_id,
            id=interface.id,
            mode=LinkMode.AUTO.value,
            subnet=subnet_id,
            force=False,
            default_gateway=False,
        )
        self.assertThat(
            interface.links,
            MatchesSetwise(
                MatchesStructure(
                    id=Equals(link_id),
                    mode=Equals(LinkMode.AUTO),
                    subnet=MatchesAll(
                        IsInstance(Subnet), MatchesStructure(id=Equals(subnet_id))
                    ),
                )
            ),
        )

    def test__interface_links_create_STATIC(self):
        origin = make_origin()
        Interface, Subnet = origin.Interface, origin.Subnet
        system_id = make_string_without_spaces()
        interface_data = {
            "system_id": system_id,
            "id": random.randint(0, 100),
            "name": make_string_without_spaces(),
            "type": InterfaceType.PHYSICAL.value,
            "links": [
                {
                    "id": random.randint(0, 100),
                    "mode": LinkMode.LINK_UP.value,
                    "subnet": {"id": random.randint(0, 100)},
                }
            ],
        }
        interface = Interface(interface_data)
        updated_data = dict(interface_data)
        link_id = random.randint(100, 200)
        subnet_id = random.randint(1, 100)
        updated_data["links"] = [
            {
                "id": link_id,
                "mode": LinkMode.STATIC.value,
                "ip_address": "192.168.122.10",
                "subnet": {"id": subnet_id},
            }
        ]
        Interface._handler.link_subnet.return_value = updated_data
        interface.links.create(
            LinkMode.STATIC,
            subnet=Subnet(subnet_id),
            ip_address="192.168.122.10",
            default_gateway=True,
            force=True,
        )
        Interface._handler.link_subnet.assert_called_once_with(
            system_id=interface.node.system_id,
            id=interface.id,
            mode=LinkMode.STATIC.value,
            subnet=subnet_id,
            ip_address="192.168.122.10",
            force=True,
            default_gateway=True,
        )
        self.assertThat(
            interface.links,
            MatchesSetwise(
                MatchesStructure(
                    id=Equals(link_id),
                    mode=Equals(LinkMode.STATIC),
                    ip_address=Equals("192.168.122.10"),
                    subnet=MatchesAll(
                        IsInstance(Subnet), MatchesStructure(id=Equals(subnet_id))
                    ),
                )
            ),
        )

    def test__interface_links_create_DHCP(self):
        origin = make_origin()
        Interface, Subnet = origin.Interface, origin.Subnet
        system_id = make_string_without_spaces()
        interface_data = {
            "system_id": system_id,
            "id": random.randint(0, 100),
            "name": make_string_without_spaces(),
            "type": InterfaceType.PHYSICAL.value,
            "links": [
                {
                    "id": random.randint(0, 100),
                    "mode": LinkMode.LINK_UP.value,
                    "subnet": {"id": random.randint(0, 100)},
                }
            ],
        }
        interface = Interface(interface_data)
        updated_data = dict(interface_data)
        link_id = random.randint(100, 200)
        subnet_id = random.randint(1, 100)
        updated_data["links"] = [
            {"id": link_id, "mode": LinkMode.DHCP.value, "subnet": {"id": subnet_id}}
        ]
        Interface._handler.link_subnet.return_value = updated_data
        interface.links.create(LinkMode.DHCP, subnet=Subnet(subnet_id))
        Interface._handler.link_subnet.assert_called_once_with(
            system_id=interface.node.system_id,
            id=interface.id,
            mode=LinkMode.DHCP.value,
            subnet=subnet_id,
            default_gateway=False,
            force=False,
        )
        self.assertThat(
            interface.links,
            MatchesSetwise(
                MatchesStructure(
                    id=Equals(link_id),
                    mode=Equals(LinkMode.DHCP),
                    subnet=MatchesAll(
                        IsInstance(Subnet), MatchesStructure(id=Equals(subnet_id))
                    ),
                )
            ),
        )

    def test__interface_links_create_LINK_UP(self):
        origin = make_origin()
        Interface, Subnet = origin.Interface, origin.Subnet
        system_id = make_string_without_spaces()
        interface_data = {
            "system_id": system_id,
            "id": random.randint(0, 100),
            "name": make_string_without_spaces(),
            "type": InterfaceType.PHYSICAL.value,
            "links": [],
        }
        interface = Interface(interface_data)
        updated_data = dict(interface_data)
        link_id = random.randint(0, 100)
        subnet_id = random.randint(1, 100)
        updated_data["links"] = [
            {"id": link_id, "mode": LinkMode.LINK_UP.value, "subnet": {"id": subnet_id}}
        ]
        Interface._handler.link_subnet.return_value = updated_data
        interface.links.create(LinkMode.LINK_UP, subnet=Subnet(subnet_id))
        Interface._handler.link_subnet.assert_called_once_with(
            system_id=interface.node.system_id,
            id=interface.id,
            mode=LinkMode.LINK_UP.value,
            subnet=subnet_id,
            default_gateway=False,
            force=False,
        )
        self.assertThat(
            interface.links,
            MatchesSetwise(
                MatchesStructure(
                    id=Equals(link_id),
                    mode=Equals(LinkMode.LINK_UP),
                    subnet=MatchesAll(
                        IsInstance(Subnet), MatchesStructure(id=Equals(subnet_id))
                    ),
                )
            ),
        )

    def test__interface_links_delete(self):
        origin = make_origin()
        Interface, Subnet = origin.Interface, origin.Subnet
        system_id = make_string_without_spaces()
        link_id = random.randint(0, 100)
        subnet_id = random.randint(1, 100)
        interface_data = {
            "system_id": system_id,
            "id": random.randint(0, 100),
            "name": make_string_without_spaces(),
            "type": InterfaceType.PHYSICAL.value,
            "links": [
                {
                    "id": link_id,
                    "mode": LinkMode.AUTO.value,
                    "subnet": {"id": subnet_id},
                }
            ],
        }
        interface = Interface(interface_data)
        updated_data = dict(interface_data)
        new_link_id = random.randint(0, 100)
        new_subnet_id = random.randint(1, 100)
        updated_data["links"] = [
            {
                "id": new_link_id,
                "mode": LinkMode.LINK_UP.value,
                "subnet": {"id": new_subnet_id},
            }
        ]
        Interface._handler.unlink_subnet.return_value = updated_data
        interface.links[0].delete()
        Interface._handler.unlink_subnet.assert_called_once_with(
            system_id=interface.node.system_id, id=interface.id, _id=link_id
        )
        self.assertThat(
            interface.links,
            MatchesSetwise(
                MatchesStructure(
                    id=Equals(new_link_id),
                    mode=Equals(LinkMode.LINK_UP),
                    subnet=MatchesAll(
                        IsInstance(Subnet), MatchesStructure(id=Equals(new_subnet_id))
                    ),
                )
            ),
        )

    def test__interface_links_set_as_default_gateway(self):
        origin = make_origin()
        Interface = origin.Interface
        system_id = make_string_without_spaces()
        link_id = random.randint(0, 100)
        subnet_id = random.randint(1, 100)
        interface_data = {
            "system_id": system_id,
            "id": random.randint(0, 100),
            "name": make_string_without_spaces(),
            "type": InterfaceType.PHYSICAL.value,
            "links": [
                {
                    "id": link_id,
                    "mode": LinkMode.AUTO.value,
                    "subnet": {"id": subnet_id},
                }
            ],
        }
        interface = Interface(interface_data)
        interface.links[0].set_as_default_gateway()
        Interface._handler.set_default_gateway.assert_called_once_with(
            system_id=interface.node.system_id, id=interface.id, link_id=link_id
        )



"""Objects for cache sets."""

__all__ = ["BcacheCacheSet", "BcacheCacheSets"]

from typing import Union

from . import ObjectSet, ObjectType
from .nodes import Node
from .block_devices import BlockDevice
from .partitions import Partition
from .filesystem_groups import DeviceField, FilesystemGroup


class BcacheCacheSetType(ObjectType):
    """Metaclass for `BcacheCacheSet`."""

    async def read(cls, node, id):
        """Get `BcacheCacheSet` by `id`."""
        if isinstance(node, str):
            system_id = node
        elif isinstance(node, Node):
            system_id = node.system_id
        else:
            raise TypeError("node must be a Node or str, not %s" % type(node).__name__)
        return cls(await cls._handler.read(system_id=system_id, id=id))


class BcacheCacheSet(FilesystemGroup, metaclass=BcacheCacheSetType):
    """A cache set on a machine."""

    cache_device = DeviceField("cache_device")

    def __repr__(self):
        return super(BcacheCacheSet, self).__repr__(fields={"name", "cache_device"})

    async def delete(self):
        """Delete this cache set."""
        await self._handler.delete(system_id=self.node.system_id, id=self.id)


class BcacheCacheSetsType(ObjectType):
    """Metaclass for `BcacheCacheSets`."""

    async def read(cls, node):
        """Get list of `BcacheCacheSet`'s for `node`."""
        if isinstance(node, str):
            system_id = node
        elif isinstance(node, Node):
            system_id = node.system_id
        else:
            raise TypeError("node must be a Node or str, not %s" % type(node).__name__)
        data = await cls._handler.read(system_id=system_id)
        return cls(
            cls._object(item, local_data={"node_system_id": system_id}) for item in data
        )

    async def create(
        cls, node: Union[Node, str], cache_device: Union[BlockDevice, Partition]
    ):
        """
        Create a BcacheCacheSet on a Node.

        :param node: Node to create the interface on.
        :type node: `Node` or `str`
        :param cache_device: Block device or partition to create
            the cache set on.
        :type cache_device: `BlockDevice` or `Partition`
        """
        params = {}
        if isinstance(node, str):
            params["system_id"] = node
        elif isinstance(node, Node):
            params["system_id"] = node.system_id
        else:
            raise TypeError(
                "node must be a Node or str, not %s" % (type(node).__name__)
            )
        if isinstance(cache_device, BlockDevice):
            params["cache_device"] = cache_device.id
        elif isinstance(cache_device, Partition):
            params["cache_partition"] = cache_device.id
        else:
            raise TypeError(
                "cache_device must be a BlockDevice or Partition, not %s"
                % (type(cache_device).__name__)
            )

        return cls._object(await cls._handler.create(**params))


class BcacheCacheSets(ObjectSet, metaclass=BcacheCacheSetsType):
    """The set of cache sets on a machine."""



"""Objects for partitions."""

__all__ = ["Partition", "Partitions"]

from . import check, Object, ObjectField, ObjectFieldRelated, ObjectSet, ObjectType
from .nodes import Node
from .block_devices import BlockDevice


def map_device_id_to_dict(instance, value):
    """Convert a device_id into a dictionary for BlockDevice."""
    return {
        "system_id": instance._data["system_id"],
        "id": value,
        "__incomplete__": True,
    }


class PartitionType(ObjectType):
    """Metaclass for `Partition`."""

    async def read(cls, node, block_device, id):
        """Get `Partition` by `id`."""
        if isinstance(node, str):
            system_id = node
        elif isinstance(node, Node):
            system_id = node.system_id
        else:
            raise TypeError("node must be a Node or str, not %s" % type(node).__name__)
        if isinstance(block_device, int):
            block_device = block_device
        elif isinstance(block_device, BlockDevice):
            block_device = block_device.id
        else:
            raise TypeError(
                "node must be a Node or str, not %s" % type(block_device).__name__
            )
        return cls(
            await cls._handler.read(system_id=system_id, device_id=block_device, id=id)
        )


class Partition(Object, metaclass=PartitionType):
    """A partition on a block device."""

    block_device = ObjectFieldRelated(
        "device_id", "BlockDevice", readonly=True, pk=0, map_func=map_device_id_to_dict
    )
    id = ObjectField.Checked("id", check(int), readonly=True, pk=1)
    uuid = ObjectField.Checked("uuid", check(str), readonly=True)
    path = ObjectField.Checked("path", check(str), readonly=True)
    size = ObjectField.Checked("size", check(int), readonly=True)
    used_for = ObjectField.Checked("used_for", check(str), readonly=True)

    filesystem = ObjectFieldRelated("filesystem", "Filesystem", readonly=True)

    def __repr__(self):
        return super(Partition, self).__repr__(fields={"path", "size"})

    async def delete(self):
        """Delete this partition."""
        await self._handler.delete(
            system_id=self.block_device.node.system_id,
            device_id=self.block_device.id,
            id=self.id,
        )

    async def format(self, fstype, *, uuid=None):
        """Format this partition."""
        self._reset(
            await self._handler.format(
                system_id=self.block_device.node.system_id,
                device_id=self.block_device.id,
                id=self.id,
                fstype=fstype,
                uuid=uuid,
            )
        )

    async def unformat(self):
        """Unformat this partition."""
        self._reset(
            await self._handler.unformat(
                system_id=self.block_device.node.system_id,
                device_id=self.block_device.id,
                id=self.id,
            )
        )

    async def mount(self, mount_point, *, mount_options=None):
        """Mount this partition."""
        self._reset(
            await self._handler.mount(
                system_id=self.block_device.node.system_id,
                device_id=self.block_device.id,
                id=self.id,
                mount_point=mount_point,
                mount_options=mount_options,
            )
        )

    async def umount(self):
        """Unmount this partition."""
        self._reset(
            await self._handler.unmount(
                system_id=self.block_device.node.system_id,
                device_id=self.block_device.id,
                id=self.id,
            )
        )


class PartitionsType(ObjectType):
    """Metaclass for `Partitions`."""

    async def read(cls, node, block_device):
        """Get list of `Partitions`'s for `node` and `block_device`."""
        if isinstance(node, str):
            system_id = node
        elif isinstance(node, Node):
            system_id = node.system_id
        else:
            raise TypeError("node must be a Node or str, not %s" % type(node).__name__)
        if isinstance(block_device, int):
            block_device = block_device
        elif isinstance(block_device, BlockDevice):
            block_device = block_device.id
        else:
            raise TypeError(
                "node must be a Node or str, not %s" % type(block_device).__name__
            )
        data = await cls._handler.read(system_id=system_id, device_id=block_device)
        return cls(cls._object(item) for item in data)

    async def create(cls, block_device: BlockDevice, size: int):
        """
        Create a partition on a block device.

        :param block_device: BlockDevice to create the paritition on.
        :type block_device: `BlockDevice`
        :param size: The size of the partition in bytes.
        :type size: `int`
        """
        params = {}
        if isinstance(block_device, BlockDevice):
            params["system_id"] = block_device.node.system_id
            params["device_id"] = block_device.id
        else:
            raise TypeError(
                "block_device must be a BlockDevice, not %s"
                % (type(block_device).__name__)
            )

        if not size:
            raise ValueError("size must be provided and greater than zero.")
        params["size"] = size
        return cls._object(await cls._handler.create(**params))


class Partitions(ObjectSet, metaclass=PartitionsType):
    """The set of partitions on a block device."""



__all__ = ["NodeStatus"]

import enum


class NodeStatus(enum.IntEnum):
    #: A node starts out as NEW (DEFAULT is an alias for NEW).
    DEFAULT = 0
    #: The node has been created and has a system ID assigned to it.
    NEW = 0
    #: Testing and other commissioning steps are taking place.
    COMMISSIONING = 1
    #: The commissioning step failed.
    FAILED_COMMISSIONING = 2
    #: The node can't be contacted.
    MISSING = 3
    #: The node is in the general pool ready to be deployed.
    READY = 4
    #: The node is ready for named deployment.
    RESERVED = 5
    #: The node has booted into the operating system of its owner's choice
    #: and is ready for use.
    DEPLOYED = 6
    #: The node has been removed from service manually until an admin
    #: overrides the retirement.
    RETIRED = 7
    #: The node is broken: a step in the node lifecyle failed.
    #: More details can be found in the node's event log.
    BROKEN = 8
    #: The node is being installed.
    DEPLOYING = 9
    #: The node has been allocated to a user and is ready for deployment.
    ALLOCATED = 10
    #: The deployment of the node failed.
    FAILED_DEPLOYMENT = 11
    #: The node is powering down after a release request.
    RELEASING = 12
    #: The releasing of the node failed.
    FAILED_RELEASING = 13
    #: The node is erasing its disks.
    DISK_ERASING = 14
    #: The node failed to erase its disks.
    FAILED_DISK_ERASING = 15
    #: The node is in rescue mode.
    RESCUE_MODE = 16
    #: The node is entering rescue mode.
    ENTERING_RESCUE_MODE = 17
    #: The node failed to enter rescue mode.
    FAILED_ENTERING_RESCUE_MODE = 18
    #: The node is exiting rescue mode.
    EXITING_RESCUE_MODE = 19
    #: The node failed to exit rescue mode.
    FAILED_EXITING_RESCUE_MODE = 20
    #: Running tests on Node
    TESTING = 21
    #: Testing has failed
    FAILED_TESTING = 22


class NodeType(enum.IntEnum):
    #: Machine
    MACHINE = 0
    #: Device
    DEVICE = 1
    #: Rack
    RACK_CONTROLLER = 2
    #: Region
    REGION_CONTROLLER = 3
    #: Region+Rack
    REGION_AND_RACK_CONTROLLER = 4


class PowerState(enum.Enum):
    #: On
    ON = "on"
    #: Off
    OFF = "off"
    #: Unknown
    UNKNOWN = "unknown"
    #: Error
    ERROR = "error"


class PowerStopMode(enum.Enum):
    #: Perform hard stop.
    HARD = "hard"
    #: Perform soft stop.
    SOFT = "soft"


class RDNSMode(enum.IntEnum):
    #: Do not generate reverse DNS for this Subnet.
    DISABLED = 0
    #: Generate reverse DNS only for the CIDR.
    ENABLED = 1
    #: Generate RFC2317 glue if needed (Subnet is too small for its own zone.)
    RFC2317 = 2


class IPRangeType(enum.Enum):
    #: Dynamic IP Range.
    DYNAMIC = "dynamic"
    #: Reserved for exclusive use by MAAS or user.
    RESERVED = "reserved"


class InterfaceType(enum.Enum):
    #: Physical interface.
    PHYSICAL = "physical"
    #: Bonded interface.
    BOND = "bond"
    #: Bridge interface.
    BRIDGE = "bridge"
    #: VLAN interface.
    VLAN = "vlan"
    #: Interface not linked to a node.
    UNKNOWN = "unknown"


class LinkMode(enum.Enum):
    #: IP is auto assigned by MAAS.
    AUTO = "auto"
    #: IP is assigned by a DHCP server.
    DHCP = "dhcp"
    #: IP is statically assigned.
    STATIC = "static"
    #: Connected to subnet with no IP address.
    LINK_UP = "link_up"


class BlockDeviceType(enum.Enum):
    #: Physical block device.
    PHYSICAL = "physical"
    #: Virtual block device.
    VIRTUAL = "virtual"


class PartitionTableType(enum.Enum):
    #: Master boot record
    MBR = "mbr"
    #: GUID Partition Table
    GPT = "gpt"


class RaidLevel(enum.Enum):
    #: RAID level 0
    RAID_0 = "raid-0"
    #: RAID level 1
    RAID_1 = "raid-1"
    #: RAID level 5
    RAID_5 = "raid-5"
    #: RAID level 6
    RAID_6 = "raid-6"
    #: RAID level 10
    RAID_10 = "raid-10"


class CacheMode(enum.Enum):
    #: Writeback
    WRITEBACK = "writeback"
    #: Writethough
    WRITETHROUGH = "writethrough"
    #: Writearound
    WRITEAROUND = "writearound"



"""Test for `maas.client.viscera.boot_source_selections`."""

import random

from testtools.matchers import Equals, MatchesStructure

from .. import boot_source_selections, boot_sources
from ...testing import make_name_without_spaces, TestCase
from ..testing import bind


def make_origin():
    # Create a new origin with BootSourceSelections and BootSourceSelection.
    # The former refers to the latter via the origin, hence why it must be
    # bound.
    return bind(
        boot_source_selections.BootSourceSelections,
        boot_source_selections.BootSourceSelection,
        boot_sources.BootSource,
    )


def make_boot_source():
    return boot_sources.BootSource(
        {
            "id": random.randint(0, 100),
            "url": "http://images.maas.io/ephemeral-v3/daily/",
            "keyring_filename": ("/usr/share/keyrings/ubuntu-cloudimage-keyring.gpg"),
            "keyring_data": "",
        }
    )


class TestBootSourceSelection(TestCase):
    def test__string_representation_includes_defined_keys(self):
        selection = boot_source_selections.BootSourceSelection(
            {
                "os": make_name_without_spaces("os"),
                "release": make_name_without_spaces("release"),
                "arches": [make_name_without_spaces("arch")],
                "subarches": [make_name_without_spaces("subarches")],
                "labels": [make_name_without_spaces("labels")],
            },
            {"boot_source_id": random.randint(0, 100)},
        )
        self.assertThat(
            repr(selection),
            Equals(
                "<BootSourceSelection arches=%(arches)r labels=%(labels)r "
                "os=%(os)r release=%(release)r subarches=%(subarches)r>"
                % (selection._data)
            ),
        )

    def test__read(self):
        source = make_boot_source()
        selection_id = random.randint(0, 100)
        os = make_name_without_spaces("os")
        release = make_name_without_spaces("release")
        arches = [make_name_without_spaces("arch")]
        subarches = [make_name_without_spaces("subarches")]
        labels = [make_name_without_spaces("labels")]

        origin = make_origin()
        BootSourceSelection = origin.BootSourceSelection
        BootSourceSelection._handler.read.return_value = {
            "id": selection_id,
            "os": os,
            "release": release,
            "arches": arches,
            "subarches": subarches,
            "labels": labels,
        }

        selection = BootSourceSelection.read(source, selection_id)
        BootSourceSelection._handler.read.assert_called_once_with(
            boot_source_id=source.id, id=selection_id
        )
        self.assertThat(
            selection,
            MatchesStructure.byEquality(
                id=selection_id,
                boot_source=origin.BootSource(
                    source.id, {"bootsourceselection": selection}
                ),
                os=os,
                release=release,
                arches=arches,
                subarches=subarches,
                labels=labels,
            ),
        )

    def test__delete(self):
        source = make_boot_source()
        selection_id = random.randint(0, 100)
        os = make_name_without_spaces("os")
        release = make_name_without_spaces("release")
        arches = [make_name_without_spaces("arch")]
        subarches = [make_name_without_spaces("subarches")]
        labels = [make_name_without_spaces("labels")]

        BootSourceSelection = make_origin().BootSourceSelection
        selection = BootSourceSelection(
            {
                "id": selection_id,
                "os": os,
                "release": release,
                "arches": arches,
                "subarches": subarches,
                "labels": labels,
            },
            {"boot_source_id": source.id},
        )

        selection.delete()
        BootSourceSelection._handler.delete.assert_called_once_with(
            boot_source_id=source.id, id=selection_id
        )


class TestBootSources(TestCase):
    def test__read_raises_TypeError_when_no_BootSource(self):
        BootSourceSelections = make_origin().BootSourceSelections
        self.assertRaises(TypeError, BootSourceSelections.read, random.randint(0, 100))

    def test__read(self):
        source = make_boot_source()
        origin = make_origin()
        BootSourceSelections = origin.BootSourceSelections
        BootSourceSelections._handler.read.return_value = [
            {"id": random.randint(0, 9)},
            {"id": random.randint(10, 19)},
        ]

        selections = BootSourceSelections.read(source)
        self.assertEquals(2, len(selections))
        self.assertEquals(
            origin.BootSource(source.id, {"bootsourceselection": selections[0]}),
            selections[0].boot_source,
        )
        self.assertEquals(
            origin.BootSource(source.id, {"bootsourceselection": selections[1]}),
            selections[1].boot_source,
        )

    def test__create_raises_TypeError_when_no_BootSource(self):
        os = make_name_without_spaces("os")
        release = make_name_without_spaces("release")

        BootSourceSelections = make_origin().BootSourceSelections

        self.assertRaises(
            TypeError, BootSourceSelections.create, random.randint(0, 100), os, release
        )

    def test__create__without_kwargs(self):
        source = make_boot_source()
        selection_id = random.randint(0, 100)
        os = make_name_without_spaces("os")
        release = make_name_without_spaces("release")

        origin = make_origin()
        BootSourceSelections = origin.BootSourceSelections
        BootSourceSelections._handler.create.return_value = {
            "id": selection_id,
            "os": os,
            "release": release,
            "arches": ["*"],
            "subarches": ["*"],
            "labels": ["*"],
        }

        selection = BootSourceSelections.create(source, os, release)
        BootSourceSelections._handler.create.assert_called_once_with(
            boot_source_id=source.id,
            os=os,
            release=release,
            arches=["*"],
            subarches=["*"],
            labels=["*"],
        )
        self.assertThat(
            selection,
            MatchesStructure.byEquality(
                id=selection_id,
                boot_source=origin.BootSource(
                    source.id, {"bootsourceselection": selection}
                ),
                os=os,
                release=release,
                arches=["*"],
                subarches=["*"],
                labels=["*"],
            ),
        )

    def test__create__with_kwargs(self):
        source = make_boot_source()
        selection_id = random.randint(0, 100)
        os = make_name_without_spaces("os")
        release = make_name_without_spaces("release")
        arches = [make_name_without_spaces("arch")]
        subarches = [make_name_without_spaces("subarches")]
        labels = [make_name_without_spaces("labels")]

        origin = make_origin()
        BootSourceSelections = origin.BootSourceSelections
        BootSourceSelections._handler.create.return_value = {
            "id": selection_id,
            "os": os,
            "release": release,
            "arches": arches,
            "subarches": subarches,
            "labels": labels,
        }

        selection = BootSourceSelections.create(
            source, os, release, arches=arches, subarches=subarches, labels=labels
        )
        BootSourceSelections._handler.create.assert_called_once_with(
            boot_source_id=source.id,
            os=os,
            release=release,
            arches=arches,
            subarches=subarches,
            labels=labels,
        )
        self.assertThat(
            selection,
            MatchesStructure.byEquality(
                id=selection_id,
                boot_source=origin.BootSource(
                    source.id, {"bootsourceselection": selection}
                ),
                os=os,
                release=release,
                arches=arches,
                subarches=subarches,
                labels=labels,
            ),
        )



"""Helpers to assemble and render tabular data."""

__all__ = ["Column", "RenderTarget", "Table"]

from abc import ABCMeta, abstractmethod
import collections
from collections.abc import Iterable
import csv
import enum
from io import StringIO
import json
from os import linesep

import colorclass
import terminaltables
import yaml


class RenderTarget(enum.Enum):

    # Human-readable tabluar output.
    plain = "plain"
    pretty = "pretty"

    # Machine-readable output.
    yaml = "yaml"
    json = "json"
    csv = "csv"

    def __str__(self):
        # Return the value. This makes it better to use as a choices option to
        # argparse at the command-line.
        return self.value


class Table(metaclass=ABCMeta):
    def __init__(self, *columns, visible_columns=None):
        super(Table, self).__init__()
        self.columns = collections.OrderedDict(
            (column.name, column) for column in columns
        )
        if visible_columns is None:
            self.visible_columns = collections.OrderedDict(self.columns.items())
        else:
            self.visible_columns = collections.OrderedDict(
                (column.name, column)
                for column in columns
                if column.name in visible_columns
            )

    def __getitem__(self, name):
        return self.columns[name]

    @abstractmethod
    def get_rows(self, target, data):
        """Get the rows for the table."""

    def _filter_rows(self, rows):
        """Filter `rows` based on the visible columns."""
        filtered_rows = []
        for row in rows:
            filtered_row = []
            for idx, name in enumerate(self.columns.keys()):
                if name in self.visible_columns:
                    filtered_row.append(row[idx])
            filtered_rows.append(filtered_row)
        return filtered_rows

    def render(self, target, data):
        """Render the table."""
        rows = self.get_rows(target, data)
        rows = self._filter_rows(rows)
        renderer = getattr(self, "_render_%s" % target.name, None)
        if renderer is None:
            raise ValueError("Cannot render %r for %s." % (self.value, target))
        else:
            return renderer(rows)

    def _flatten_columns(self, columns):
        cols = []
        for _, column in columns.items():
            if isinstance(column, NestedTableColumn):
                cols.extend(self._flatten_columns(column.get_columns()))
            else:
                cols.append(column)
        return cols

    def _compute_rows(self, target, data, duplicate=False):
        columns = self.visible_columns.values()
        computed = []
        for row_data in data:
            row = []
            rows = [row]
            for datum, column in zip(row_data, columns):
                if isinstance(column, NestedTableColumn):
                    nested_rows = column.get_rows(target, datum, duplicate=duplicate)
                    orig_row = list(row)
                    row.extend(nested_rows[0])
                    for nested_row in nested_rows[1:]:
                        new_row = list(orig_row)
                        if not duplicate:
                            for idx in range(len(new_row)):
                                new_row[idx] = ""
                        new_row.extend(nested_row)
                        rows.append(new_row)
                else:
                    row.append(column.render(target, datum))
            computed.extend(rows)
        return computed

    def _render_plain(self, data):
        rows = self._compute_rows(RenderTarget.plain, data)
        rows.insert(
            0, [column.title for column in self._flatten_columns(self.visible_columns)]
        )
        return terminaltables.AsciiTable(rows).table

    def _render_pretty(self, data):
        rows = self._compute_rows(RenderTarget.pretty, data)
        rows.insert(
            0, [column.title for column in self._flatten_columns(self.visible_columns)]
        )
        return terminaltables.SingleTable(rows).table

    def _render_yaml(self, data):
        columns = self.visible_columns.values()
        rows = [
            [
                column.render(RenderTarget.yaml, datum)
                for datum, column in zip(row, columns)
            ]
            for row in data
        ]
        return yaml.safe_dump(
            {
                "columns": [
                    {"name": column.name, "title": column.title} for column in columns
                ],
                "data": [
                    {column.name: datum for column, datum in zip(columns, row)}
                    for row in rows
                ],
            },
            default_flow_style=False,
        ).rstrip(linesep)

    def _render_json(self, data):
        columns = self.visible_columns.values()
        rows = [
            [
                column.render(RenderTarget.json, datum)
                for datum, column in zip(row, columns)
            ]
            for row in data
        ]
        return json.dumps(
            {
                "columns": [
                    {"name": column.name, "title": column.title} for column in columns
                ],
                "data": [
                    {column.name: datum for column, datum in zip(columns, row)}
                    for row in rows
                ],
            }
        )

    def _render_csv(self, data):
        output = StringIO()
        writer = csv.writer(output)
        writer.writerow(
            [column.name for column in self._flatten_columns(self.visible_columns)]
        )
        writer.writerows(self._compute_rows(RenderTarget.csv, data, duplicate=True))
        return output.getvalue().rstrip(linesep)

    def __repr__(self):
        return "<%s [%s]>" % (self.__class__.__name__, " ".join(self.visible_columns))


class DetailTable(Table):
    def _filter_rows(self, rows, visible_columns=None):
        """Filter `rows` based on the visible columns."""
        if visible_columns is None:
            visible_columns = self.visible_columns
        filtered_row = []
        for idx, name in enumerate(self.columns.keys()):
            if name in self.visible_columns:
                filtered_row.append(rows[idx])
        return filtered_row

    def render(self, target, data):
        renderer = getattr(self, "_render_%s" % target.name, None)
        if renderer is None:
            raise ValueError("Cannot render %r for %s." % (self.value, target))
        else:
            return renderer(data)

    def _split_nested_tables(self):
        table_columns = collections.OrderedDict(
            (name, column)
            for name, column in self.visible_columns.items()
            if isinstance(column, NestedTableColumn)
        )
        data_columns = collections.OrderedDict(
            (name, column)
            for name, column in self.visible_columns.items()
            if not isinstance(column, NestedTableColumn)
        )
        return data_columns, table_columns

    def _render_nested_table(self, target, data, column):
        data_idx = list(self.columns.keys()).index(column.name)
        data = data[data_idx]
        table = column.get_table()
        return table.render(target, data)

    def _render_nested_tables(self, target, data, columns):
        tables = [
            self._render_nested_table(target, data, column)
            for _, column in columns.items()
        ]
        if len(tables) > 0:
            return "\n" + "\n".join(tables)
        else:
            return ""

    def _render_table(self, target, terminaltable, data):
        columns, table_columns = self._split_nested_tables()
        all_rows = self.get_rows(target, data)
        rows = self._filter_rows(all_rows, visible_columns=columns)
        rows = [
            column.render(target, datum)
            for column, datum in zip(columns.values(), rows)
        ]
        table = terminaltable(
            [(column.title, datum) for column, datum in zip(columns.values(), rows)]
        )
        table.inner_heading_row_border = False
        return table.table + self._render_nested_tables(target, all_rows, table_columns)

    def _render_plain(self, data):
        return self._render_table(RenderTarget.plain, terminaltables.AsciiTable, data)

    def _render_pretty(self, data):
        return self._render_table(RenderTarget.pretty, terminaltables.SingleTable, data)

    def _render_yaml(self, data):
        columns = self.visible_columns.values()
        rows = self.get_rows(RenderTarget.yaml, data)
        rows = self._filter_rows(rows)
        rows = [
            column.render(RenderTarget.yaml, datum)
            for column, datum in zip(columns, rows)
        ]
        return yaml.safe_dump(
            {column.name: datum for column, datum in zip(columns, rows)},
            default_flow_style=False,
        ).rstrip(linesep)

    def _render_json(self, data):
        columns = self.visible_columns.values()
        rows = self.get_rows(RenderTarget.json, data)
        rows = self._filter_rows(rows)
        rows = [
            column.render(RenderTarget.json, datum)
            for column, datum in zip(columns, rows)
        ]
        return json.dumps({column.name: datum for column, datum in zip(columns, rows)})

    def _render_csv(self, data):
        columns, table_columns = self._split_nested_tables()
        all_rows = self.get_rows(RenderTarget.csv, data)
        rows = self._filter_rows(all_rows, visible_columns=columns)
        rows = [
            column.render(RenderTarget.csv, datum)
            for column, datum in zip(columns.values(), rows)
        ]
        output = StringIO()
        writer = csv.writer(output)
        writer.writerows(
            [(column.name, datum) for column, datum in zip(columns.values(), rows)]
        )
        return output.getvalue().rstrip(linesep) + (
            self._render_nested_tables(RenderTarget.csv, all_rows, table_columns)
        )


class Column:
    def __init__(self, name, title=None):
        super(Column, self).__init__()
        self.name = name
        self.title = name if title is None else title

    def render(self, target, datum):
        if target is RenderTarget.yaml:
            return datum
        elif target is RenderTarget.json:
            return datum
        elif target is RenderTarget.csv:
            if isinstance(datum, Iterable) and not isinstance(datum, (str, bytes)):
                return ",".join(datum)
            else:
                return datum
        elif target is RenderTarget.plain:
            if datum is None:
                return ""
            elif isinstance(datum, colorclass.Color):
                return datum.value_no_colors
            elif isinstance(datum, Iterable) and not isinstance(datum, (str, bytes)):
                return "\n".join(datum)
            else:
                return str(datum)
        elif target is RenderTarget.pretty:
            if datum is None:
                return ""
            elif isinstance(datum, colorclass.Color):
                return datum
            elif isinstance(datum, Iterable) and not isinstance(datum, (str, bytes)):
                return "\n".join(datum)
            else:
                return str(datum)
        else:
            raise ValueError("Cannot render %r for %s" % (datum, target))

    def __repr__(self):
        return "<%s name=%s title=%r>" % (
            self.__class__.__name__,
            self.name,
            self.title,
        )


class NestedTableColumn(Column):
    def __init__(
        self, name, title=None, table=None, table_args=None, table_kwargs=None
    ):
        super(NestedTableColumn, self).__init__(name, title=title)
        self.table = table
        self.table_args = table_args
        self.table_kwargs = table_kwargs
        if table is None:
            raise ValueError("table is required.")

    def get_table(self):
        table_args = self.table_args
        if table_args is None:
            table_args = []
        table_kwargs = self.table_kwargs
        if table_kwargs is None:
            table_kwargs = {}
        return self.table(*table_args, **table_kwargs)

    def get_columns(self):
        return self.get_table().visible_columns

    def get_rows(self, target, data, duplicate=False):
        table = self.get_table()
        rows = table.get_rows(target, data)
        rows = table._filter_rows(rows)
        rows = table._compute_rows(target, rows, duplicate=duplicate)
        if len(rows) == 0:
            # Nested table column must always return one row even if its
            # an empty row.
            rows = [[" " for _ in range(len(table.visible_columns))]]
        return rows

    def render(self, target, datum):
        table = self.get_table()
        if target is RenderTarget.yaml:
            return yaml.safe_load(table.render(target, datum))
        elif target is RenderTarget.json:
            return json.loads(table.render(target, datum))
        else:
            raise ValueError(
                "Should not be called on a nested table column, the "
                "table render should handle this correctly."
            )



"""Objects for interfaces."""

__all__ = ["Interface", "Interfaces"]

import copy
from typing import Iterable, Union

from . import (
    check,
    Object,
    ObjectField,
    ObjectFieldRelated,
    ObjectFieldRelatedSet,
    ObjectSet,
    ObjectType,
    to,
)
from .nodes import Node
from .subnets import Subnet
from .vlans import Vlan
from ..enum import InterfaceType, LinkMode
from ..utils.diff import calculate_dict_diff


def gen_parents(parents):
    """Generate the parents to send to the handler."""
    for idx, parent in enumerate(parents):
        if isinstance(parent, Interface):
            parent = parent.id
        elif isinstance(parent, int):
            pass
        else:
            raise TypeError(
                "parent[%d] must be an Interface or int, not %s"
                % (idx, type(parent).__name__)
            )
        yield parent


def get_parent(parent):
    """Get the parent to send to the handler."""
    if isinstance(parent, Interface):
        return parent.id
    elif isinstance(parent, int):
        return parent
    else:
        raise TypeError(
            "parent must be an Interface or int, not %s" % (type(parent).__name__)
        )


class InterfaceTypeMeta(ObjectType):
    """Metaclass for `Interface`."""

    async def read(cls, node, id):
        """Get `Interface` by `id`."""
        if isinstance(node, str):
            system_id = node
        elif isinstance(node, Node):
            system_id = node.system_id
        else:
            raise TypeError("node must be a Node or str, not %s" % type(node).__name__)
        return cls(await cls._handler.read(system_id=system_id, id=id))


def map_nic_name_to_dict(instance, value):
    """Convert a name of interface into a dictionary.

    `parents` and `children` just hold a list of interface names. This is need
    so instead they can return a `ObjectSet`.

    '__incomplete__' is set so the object knows that the data passed is
    incomplete data.
    """
    return {
        "system_id": instance._data["system_id"],
        "name": value,
        "__incomplete__": True,
    }


class Interface(Object, metaclass=InterfaceTypeMeta):
    """A interface on a machine."""

    node = ObjectFieldRelated("system_id", "Node", readonly=True, pk=0)
    id = ObjectField.Checked("id", check(int), readonly=True, pk=1)
    type = ObjectField.Checked("type", to(InterfaceType), readonly=True)
    name = ObjectField.Checked("name", check(str), check(str), alt_pk=1)
    mac_address = ObjectField.Checked("mac_address", check(str), check(str))
    enabled = ObjectField.Checked("enabled", check(bool), check(bool))
    effective_mtu = ObjectField.Checked("effective_mtu", check(int), readonly=True)
    tags = ObjectField.Checked("tags", check(list), check(list))
    params = ObjectField.Checked("params", check((dict, str)), check((dict, str)))
    parents = ObjectFieldRelatedSet(
        "parents", "Interfaces", reverse=None, map_func=map_nic_name_to_dict
    )
    children = ObjectFieldRelatedSet(
        "children", "Interfaces", reverse=None, map_func=map_nic_name_to_dict
    )
    vlan = ObjectFieldRelated("vlan", "Vlan", reverse=None, use_data_setter=True)
    links = ObjectFieldRelatedSet("links", "InterfaceLinks", reverse="interface")
    discovered = ObjectFieldRelatedSet(
        "discovered", "InterfaceDiscoveredLinks", reverse=None
    )

    def __repr__(self):
        return super(Interface, self).__repr__(fields={"name", "mac_address", "type"})

    async def save(self):
        """Save this interface."""
        if set(self.tags) != set(self._orig_data["tags"]):
            self._changed_data["tags"] = ",".join(self.tags)
        elif "tags" in self._changed_data:
            del self._changed_data["tags"]
        orig_params = self._orig_data["params"]
        if not isinstance(orig_params, dict):
            orig_params = {}
        params = self.params
        if not isinstance(params, dict):
            params = {}
        self._changed_data.pop("params", None)
        self._changed_data.update(calculate_dict_diff(orig_params, params))
        if "vlan" in self._changed_data and self._changed_data["vlan"]:
            # Update uses the ID of the VLAN, not the VLAN object.
            self._changed_data["vlan"] = self._changed_data["vlan"]["id"]
            if (
                self._orig_data["vlan"]
                and "id" in self._orig_data["vlan"]
                and self._changed_data["vlan"] == (self._orig_data["vlan"]["id"])
            ):
                # VLAN didn't really change, the object was just set to the
                # same VLAN.
                del self._changed_data["vlan"]
        await super(Interface, self).save()

    async def delete(self):
        """Delete this interface."""
        await self._handler.delete(system_id=self.node.system_id, id=self.id)

    async def disconnect(self):
        """Disconnect this interface."""
        self._reset(
            await self._handler.disconnect(system_id=self.node.system_id, id=self.id)
        )


class InterfaceDiscoveredLink(Object):
    """Discovered link information on an `Interface`."""

    ip_address = ObjectField.Checked(
        "ip_address", check(str), readonly=True, default=None
    )
    subnet = ObjectFieldRelated("subnet", "Subnet", readonly=True, default=None)


class InterfaceDiscoveredLinks(ObjectSet):
    """A set of discovered links on an `Interface`."""


class InterfaceLink(Object):
    """A link on an `Interface`."""

    id = ObjectField.Checked("id", check(int), readonly=True)
    mode = ObjectField.Checked("mode", to(LinkMode), readonly=True)
    subnet = ObjectFieldRelated("subnet", "Subnet", readonly=True, default=None)
    ip_address = ObjectField.Checked(
        "ip_address", check(str), readonly=True, default=None
    )

    def __repr__(self):
        return super(InterfaceLink, self).__repr__(
            fields={"mode", "ip_address", "subnet"}
        )

    async def delete(self):
        """Delete this interface link."""
        interface = self._data["interface"]
        data = await interface._handler.unlink_subnet(
            system_id=interface.node.system_id, id=interface.id, _id=self.id
        )
        interface._data["links"] = list(data["links"])
        interface._orig_data["links"] = copy.deepcopy(interface._data["links"])

    async def set_as_default_gateway(self):
        """Set this link as the default gateway for the node."""
        interface = self._data["interface"]
        await interface._handler.set_default_gateway(
            system_id=interface.node.system_id, id=interface.id, link_id=self.id
        )


class InterfaceLinksType(ObjectType):
    """Metaclass for `InterfaceLinks`."""

    async def create(
        cls,
        interface: Interface,
        mode: LinkMode,
        subnet: Union[Subnet, int] = None,
        ip_address: str = None,
        force: bool = False,
        default_gateway: bool = False,
    ):
        """
        Create a link on `Interface` in MAAS.

        :param interface: Interface to create the link on.
        :type interface: `Interface`
        :param mode: Mode of the link.
        :type mode: `LinkMode`
        :param subnet: The subnet to create the link on (optional).
        :type subnet: `Subnet` or `int`
        :param ip_address: The IP address to assign to the link.
        :type ip_address: `str`
        :param force: If True, allows `LinkMode.LINK_UP` to be created even if
            other links already exist. Also allows the selection of any
            subnet no matter the VLAN the subnet belongs to. Using this option
            will cause all other interface links to be deleted (optional).
        :type force: `bool`
        :param default_gateway: If True, sets the gateway IP address for the
            subnet as the default gateway for the node this interface belongs
            to. Option can only be used with the `LinkMode.AUTO` and
            `LinkMode.STATIC` modes.
        :type default_gateway: `bool`

        :returns: The created InterfaceLink.
        :rtype: `InterfaceLink`
        """
        if not isinstance(interface, Interface):
            raise TypeError(
                "interface must be an Interface, not %s" % type(interface).__name__
            )
        if not isinstance(mode, LinkMode):
            raise TypeError("mode must be a LinkMode, not %s" % type(mode).__name__)
        if subnet is not None:
            if isinstance(subnet, Subnet):
                subnet = subnet.id
            elif isinstance(subnet, int):
                pass
            else:
                raise TypeError(
                    "subnet must be a Subnet or int, not %s" % type(subnet).__name__
                )
        if mode in [LinkMode.AUTO, LinkMode.STATIC]:
            if subnet is None:
                raise ValueError("subnet is required for %s" % mode)
        if default_gateway and mode not in [LinkMode.AUTO, LinkMode.STATIC]:
            raise ValueError("cannot set as default_gateway for %s" % mode)
        params = {
            "system_id": interface.node.system_id,
            "id": interface.id,
            "mode": mode.value,
            "force": force,
            "default_gateway": default_gateway,
        }
        if subnet is not None:
            params["subnet"] = subnet
        if ip_address is not None:
            params["ip_address"] = ip_address
        # The API doesn't return just the link it returns the whole interface.
        # Store the link ids before the save to find the addition at the end.
        link_ids = {link.id for link in interface.links}
        data = await interface._handler.link_subnet(**params)
        # Update the links on the interface, except for the newly created link
        # the `ManagedCreate` wrapper will add that to the interfaces link data
        # automatically.
        new_links = {link["id"]: link for link in data["links"]}
        links_diff = list(set(new_links.keys()) - link_ids)
        new_link = new_links.pop(links_diff[0])
        interface._data["links"] = list(new_links.values())
        interface._orig_data["links"] = copy.deepcopy(interface._data["links"])
        return cls._object(new_link)


class InterfaceLinks(ObjectSet, metaclass=InterfaceLinksType):
    """A set of links on an `Interface`."""


class InterfacesType(ObjectType):
    """Metaclass for `Interfaces`."""

    async def read(cls, node):
        """Get list of `Interface`'s for `node`."""
        if isinstance(node, str):
            system_id = node
        elif isinstance(node, Node):
            system_id = node.system_id
        else:
            raise TypeError("node must be a Node or str, not %s" % type(node).__name__)
        data = await cls._handler.read(system_id=system_id)
        return cls(
            cls._object(item, local_data={"node_system_id": system_id}) for item in data
        )

    async def create(
        cls,
        node: Union[Node, str],
        interface_type: InterfaceType = InterfaceType.PHYSICAL,
        *,
        name: str = None,
        mac_address: str = None,
        tags: Iterable[str] = None,
        vlan: Union[Vlan, int] = None,
        parent: Union[Interface, int] = None,
        parents: Iterable[Union[Interface, int]] = None,
        mtu: int = None,
        accept_ra: bool = None,
        autoconf: bool = None,
        bond_mode: str = None,
        bond_miimon: int = None,
        bond_downdelay: int = None,
        bond_updelay: int = None,
        bond_lacp_rate: str = None,
        bond_xmit_hash_policy: str = None,
        bridge_stp: bool = None,
        bridge_fd: int = None
    ):
        """
        Create a `Interface` in MAAS.

        :param node: Node to create the interface on.
        :type node: `Node` or `str`
        :param interface_type: Type of interface to create (optional).
        :type interface_type: `InterfaceType`
        :param name: The name for the interface (optional).
        :type name: `str`
        :param tags: List of tags to add to the interface.
        :type tags: sequence of `str`
        :param mtu: The MTU for the interface (optional).
        :type mtu: `int`
        :param vlan: VLAN the interface is connected to (optional).
        :type vlan: `Vlan` or `int`
        :param accept_ra: True if the interface should accepted router
            advertisements. (optional)
        :type accept_ra: `bool`
        :param autoconf: True if the interface should auto configure.
        :type autoconf: `bool`

        Following parameters specific to physical interface.

        :param mac_address: The MAC address for the interface.
        :type mac_address: `str`

        Following parameters specific to a bond interface.

        :param parents: Parent interfaces that make up the bond.
        :type parents: sequence of `Interface` or `int`
        :param mac_address: MAC address to use for the bond (optional).
        :type mac_address: `str`
        :param bond_mode: The operating mode of the bond (optional).
        :type bond_mode: `str`
        :param bond_miimon: The link monitoring freqeuncy in
            milliseconds (optional).
        :type bond_miimon: `int`
        :param bond_downdelay: Specifies the time, in milliseconds, to wait
            before disabling a slave after a link failure has been detected
            (optional).
        :type bond_downdelay: `int`
        :param bond_updelay: Specifies the time, in milliseconds, to wait
            before enabling a slave after a link recovery has been detected.
        :type bond_updelay: `int`
        :param bond_lacp_rate: Option specifying the rate in which we'll ask
            our link partner to transmit LACPDU packets in 802.3ad
            mode (optional).
        :type bond_lacp_rate: `str`
        :param bond_xmit_hash_policy: The transmit hash policy to use for
            slave selection in balance-xor, 802.3ad, and tlb modes(optional).
        :type bond_xmit_hash_policy: `str`

        Following parameters specific to a VLAN interface.

        :param parent: Parent interface for this VLAN interface.
        :type parent: `Interface` or `int`

        Following parameters specific to a Bridge interface.

        :param parent: Parent interface for this bridge interface.
        :type parent: `Interface` or `int`
        :param mac_address: The MAC address for the interface (optional).
        :type mac_address: `str`
        :param bridge_stp: Turn spanning tree protocol on or off (optional).
        :type bridge_stp: `bool`
        :param bridge_fd: Set bridge forward delay to time seconds (optional).
        :type bridge_fd: `int`

        :returns: The created Interface.
        :rtype: `Interface`
        """
        params = {}
        if isinstance(node, str):
            params["system_id"] = node
        elif isinstance(node, Node):
            params["system_id"] = node.system_id
        else:
            raise TypeError(
                "node must be a Node or str, not %s" % (type(node).__name__)
            )

        if name is not None:
            params["name"] = name
        if tags is not None:
            params["tags"] = tags
        if mtu is not None:
            params["mtu"] = mtu
        if vlan is not None:
            if isinstance(vlan, Vlan):
                params["vlan"] = vlan.id
            elif isinstance(vlan, int):
                params["vlan"] = vlan
            else:
                raise TypeError(
                    "vlan must be a Vlan or int, not %s" % (type(vlan).__name__)
                )
        if accept_ra is not None:
            params["accept_ra"] = accept_ra
        if autoconf is not None:
            params["autoconf"] = autoconf

        handler = None
        if not isinstance(interface_type, InterfaceType):
            raise TypeError(
                "interface_type must be an InterfaceType, not %s"
                % (type(interface_type).__name__)
            )
        if interface_type == InterfaceType.PHYSICAL:
            handler = cls._handler.create_physical
            if mac_address:
                params["mac_address"] = mac_address
            else:
                raise ValueError("mac_address required for physical interface")
        elif interface_type == InterfaceType.BOND:
            handler = cls._handler.create_bond
            if parent is not None:
                raise ValueError("use parents not parent for bond interface")
            if not isinstance(parents, Iterable):
                raise TypeError(
                    "parents must be a iterable, not %s" % (type(parents).__name__)
                )
            if len(parents) == 0:
                raise ValueError("at least one parent required for bond interface")
            params["parents"] = list(gen_parents(parents))
            if not name:
                raise ValueError("name is required for bond interface")
            if mac_address is not None:
                params["mac_address"] = mac_address
            if bond_mode is not None:
                params["bond_mode"] = bond_mode
            if bond_miimon is not None:
                params["bond_miimon"] = bond_miimon
            if bond_downdelay is not None:
                params["bond_downdelay"] = bond_downdelay
            if bond_updelay is not None:
                params["bond_updelay"] = bond_updelay
            if bond_lacp_rate is not None:
                params["bond_lacp_rate"] = bond_lacp_rate
            if bond_xmit_hash_policy is not None:
                params["bond_xmit_hash_policy"] = bond_xmit_hash_policy
        elif interface_type == InterfaceType.VLAN:
            handler = cls._handler.create_vlan
            if parents is not None:
                raise ValueError("use parent not parents for VLAN interface")
            if parent is None:
                raise ValueError("parent is required for VLAN interface")
            params["parent"] = get_parent(parent)
            if vlan is None:
                raise ValueError("vlan is required for VLAN interface")
        elif interface_type == InterfaceType.BRIDGE:
            handler = cls._handler.create_bridge
            if parents is not None:
                raise ValueError("use parent not parents for bridge interface")
            if parent is None:
                raise ValueError("parent is required for bridge interface")
            params["parent"] = get_parent(parent)
            if not name:
                raise ValueError("name is required for bridge interface")
            if mac_address is not None:
                params["mac_address"] = mac_address
            if bridge_stp is not None:
                params["bridge_stp"] = bridge_stp
            if bridge_fd is not None:
                params["bridge_fd"] = bridge_fd
        else:
            raise ValueError("cannot create an interface of type: %s" % interface_type)

        return cls._object(await handler(**params))


class Interfaces(ObjectSet, metaclass=InterfacesType):
    """The set of interfaces on a machine."""

    @property
    def by_name(self):
        """Return mapping of name of interface to `Interface`."""
        return {interface.name: interface for interface in self}

    def get_by_name(self, name):
        """Return an `Interface` by its name."""
        return self.by_name[name]



"""Test for `maas.client.viscera.logical_volumes`."""

import random
from testtools.matchers import Equals, IsInstance

from .. import block_devices, logical_volumes, nodes, volume_groups
from ...testing import make_name_without_spaces, TestCase
from ..testing import bind


def make_origin():
    # Create a new origin with Devices and Device. The former refers to the
    # latter via the origin, hence why it must be bound.
    return bind(
        logical_volumes.LogicalVolume,
        logical_volumes.LogicalVolumes,
        block_devices.BlockDevice,
        block_devices.BlockDevices,
        nodes.Node,
        volume_groups.VolumeGroup,
    )


class TestLogicalVolume(TestCase):
    def test__string_representation_includes_only_name(self):
        volume = logical_volumes.LogicalVolume(
            {"name": make_name_without_spaces("name")}
        )
        self.assertThat(
            repr(volume), Equals("<LogicalVolume name=%(name)r>" % volume._data)
        )


class TestLogicalVolumes(TestCase):
    def test__create(self):
        origin = make_origin()

        system_id = make_name_without_spaces("system-id")
        lv_id = random.randint(1, 20)
        lv_name = make_name_without_spaces("lvname")

        VolumeGroup = origin.VolumeGroup
        vg = VolumeGroup({"system_id": system_id, "id": random.randint(21, 30)})
        VolumeGroup._handler.create_logical_volume.return_value = {
            "system_id": system_id,
            "id": lv_id,
        }

        uuid = make_name_without_spaces("uuid")
        tags = [make_name_without_spaces("tag")]

        BlockDevice = origin.BlockDevice
        BlockDevice._handler.read.return_value = {
            "system_id": system_id,
            "id": lv_id,
            "name": lv_name,
            "tags": [],
            "uuid": uuid,
        }

        LogicalVolumes = origin.LogicalVolumes
        observed = LogicalVolumes.create(vg, lv_name, 10 * 1024, uuid=uuid, tags=tags)
        self.assertThat(observed, IsInstance(logical_volumes.LogicalVolume))
        self.assertThat(observed.name, Equals(lv_name))

        VolumeGroup._handler.create_logical_volume.assert_called_once_with(
            system_id=system_id, id=vg.id, name=lv_name, size=10 * 1024, uuid=uuid
        )
        BlockDevice._handler.read.assert_called_once_with(system_id=system_id, id=lv_id)
        BlockDevice._handler.add_tag.assert_called_once_with(
            system_id=system_id, id=lv_id, tag=tags[0]
        )



# Copyright 2016-2017 Canonical Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Utilities for the MAAS client."""

__all__ = [
    "api_url",
    "coalesce",
    "get_all_subclasses",
    "parse_docstring",
    "prepare_payload",
    "retries",
    "sign",
    "Spinner",
    "vars_class",
]


from collections import namedtuple
from collections.abc import Iterable
from functools import lru_cache, partial
from inspect import cleandoc, getdoc
from itertools import chain, cycle, repeat
import re
import sys
import threading
from time import time
from urllib.parse import quote_plus, urlparse

from oauthlib import oauth1

from .multipart import build_multipart_message, encode_multipart_message


def urlencode(data):
    """A version of `urllib.urlencode` that isn't insane.

    This only cares that `data` is an iterable of iterables. Each sub-iterable
    must be of overall length 2, i.e. a name/value pair.

    Unicode strings will be encoded to UTF-8. This is what Django expects; see
    `smart_text` in the Django documentation.
    """

    def dec(string):
        if isinstance(string, bytes):
            string = string.decode("utf-8")
        return quote_plus(string)

    return "&".join("%s=%s" % (dec(name), dec(value)) for name, value in data)


def prepare_payload(op, method, uri, data):
    """Return the URI (modified perhaps) and body and headers.

    - For GET requests, encode parameters in the query string.

    - Otherwise always encode parameters in the request body.

    - Except op; this can always go in the query string.

    :param method: The HTTP method.
    :param uri: The URI of the action.
    :param data: An iterable of ``name, value`` or ``name, opener``
        tuples (see `name_value_pair`) to pack into the body or
        query, depending on the type of request.
    """
    query = [] if op is None else [("op", op)]

    def slurp(opener):
        with opener() as fd:
            return fd.read()

    if method == "GET":
        headers, body = [], None
        query.extend(
            (name, slurp(value) if callable(value) else value) for name, value in data
        )
    else:
        # Even if data is empty, construct a multipart request body. Piston
        # (server-side) sets `request.data` to `None` if there's no payload.
        message = build_multipart_message(data)
        headers, body = encode_multipart_message(message)

    uri = urlparse(uri)._replace(query=urlencode(query)).geturl()
    return uri, body, headers


class OAuthSigner:
    """Helper class to OAuth-sign an HTTP request."""

    def __init__(
        self, token_key, token_secret, consumer_key, consumer_secret, realm="OAuth"
    ):
        """Initialize a ``OAuthAuthorizer``.

        :type token_key: Unicode string.
        :type token_secret: Unicode string.
        :type consumer_key: Unicode string.
        :type consumer_secret: Unicode string.

        :param realm: Optional.
        """

        def _to_unicode(string):
            if isinstance(string, bytes):
                return string.decode("ascii")
            else:
                return string

        self.token_key = _to_unicode(token_key)
        self.token_secret = _to_unicode(token_secret)
        self.consumer_key = _to_unicode(consumer_key)
        self.consumer_secret = _to_unicode(consumer_secret)
        self.realm = _to_unicode(realm)

    def sign_request(self, url, method, body, headers):
        """Sign a request.

        :param url: The URL to which the request is to be sent.
        :param headers: The headers in the request. These will be updated with
            the signature.
        """
        # The use of PLAINTEXT here was copied from MAAS, but we should switch
        # to HMAC once it works server-side.
        client = oauth1.Client(
            self.consumer_key,
            self.consumer_secret,
            self.token_key,
            self.token_secret,
            signature_method=oauth1.SIGNATURE_PLAINTEXT,
            realm=self.realm,
        )
        # To preserve API backward compatibility convert an empty string body
        # to `None`. The old "oauth" library would treat the empty string as
        # "no body", but "oauthlib" requires `None`.
        body = None if body is None or len(body) == 0 else body
        uri, signed_headers, body = client.sign(url, method, body, headers)
        headers.update(signed_headers)


def sign(uri, headers, credentials):
    """Sign the URI and headers.

    A request method of `GET` with no body content is assumed.

    :param credentials: A tuple of consumer key, token key, and token secret.
    """
    consumer_key, token_key, token_secret = credentials
    auth = OAuthSigner(token_key, token_secret, consumer_key, "")
    auth.sign_request(uri, method="GET", body=None, headers=headers)


re_paragraph_splitter = re.compile(r"(?:\r\n){2,}|\r{2,}|\n{2,}", re.MULTILINE)

paragraph_split = re_paragraph_splitter.split
docstring_split = partial(paragraph_split, maxsplit=1)
remove_line_breaks = lambda string: (
    " ".join(line.strip() for line in string.splitlines())
)

newline = "\n"
empty = ""


docstring = namedtuple("docstring", ("title", "body"))


@lru_cache(2**10)
def parse_docstring(thing):
    """Parse a Python docstring, or the docstring found on `thing`.

    :return: a ``(title, body)`` tuple. As per docstring convention, title is
        the docstring's first paragraph and body is the rest.
    """
    assert not isinstance(thing, bytes)
    doc = cleandoc(thing) if isinstance(thing, str) else getdoc(thing)
    doc = empty if doc is None else doc
    assert not isinstance(doc, bytes)
    # Break the docstring into two parts: title and body.
    parts = docstring_split(doc)
    if len(parts) == 2:
        title, body = parts[0], parts[1]
    else:
        title, body = parts[0], empty
    # Remove line breaks from the title line.
    title = remove_line_breaks(title)
    # Normalise line-breaks on newline.
    body = body.replace("\r\n", newline).replace("\r", newline)
    return docstring(title, body)


def ensure_trailing_slash(string):
    """Ensure that `string` has a trailing forward-slash."""
    return (string + "/") if not string.endswith("/") else string


def api_url(string):
    """Ensure that `string` looks like a URL to the API.

    This ensures that the API version is specified explicitly (i.e. the path
    ends with /api/{version}). If not, version 2.0 is selected. It also
    ensures that the path ends with a forward-slash.

    This is suitable for use as an argument type with argparse.
    """
    url = urlparse(string)
    url = url._replace(path=ensure_trailing_slash(url.path))
    if re.search("/api/[0-9.]+/?$", url.path) is None:
        url = url._replace(path=url.path + "api/2.0/")
    return url.geturl()


def get_all_subclasses(cls):
    """Get all subclasses of `cls`, recursively."""
    for cls in cls.__subclasses__():
        yield from get_all_subclasses(cls)
        yield cls


def vars_class(cls):
    """Return a dict of vars for the given class, including all ancestors.

    This differs from the usual behaviour of `vars` which returns attributes
    belonging to the given class and not its ancestors.
    """
    return dict(chain.from_iterable(vars(cls).items() for cls in reversed(cls.__mro__)))


def retries(timeout=30, intervals=1, time=time):
    """Helper for retrying something, sleeping between attempts.

    Returns a generator that yields ``(elapsed, remaining, wait)`` tuples,
    giving times in seconds. The last item, `wait`, is the suggested amount of
    time to sleep before trying again.

    :param timeout: From now, how long to keep iterating, in seconds. This can
        be specified as a number, or as an iterable. In the latter case, the
        iterator is advanced each time an interval is needed. This allows for
        back-off strategies.
    :param intervals: The sleep between each iteration, in seconds, an an
        iterable from which to obtain intervals.
    :param time: A callable that returns the current time in seconds.
    """
    start = time()
    end = start + timeout

    if isinstance(intervals, Iterable):
        intervals = iter(intervals)
    else:
        intervals = repeat(intervals)

    return gen_retries(start, end, intervals, time)


def gen_retries(start, end, intervals, time=time):
    """Helper for retrying something, sleeping between attempts.

    Yields ``(elapsed, remaining, wait)`` tuples, giving times in seconds. The
    last item, `wait`, is the suggested amount of time to sleep before trying
    again.

    This function works in concert with `retries`. It's split out so that
    `retries` can capture the correct start time rather than the time at which
    it is first iterated.

    :param start: The start time, in seconds, of this generator. This must be
        congruent with the `IReactorTime` argument passed to this generator.
    :param end: The desired end time, in seconds, of this generator. This must
        be congruent with the `IReactorTime` argument passed to this
        generator.
    :param intervals: A iterable of intervals, each in seconds, which should
        be used as hints for the `wait` value that's generated.
    :param time: A callable that returns the current time in seconds.
    """
    for interval in intervals:
        now = time()
        if now < end:
            wait = min(interval, end - now)
            yield now - start, end - now, wait
        else:
            yield now - start, end - now, 0
            break


def coalesce(*values, default=None):
    """Return the first argument that is not `None`.

    If all arguments are `None`, return `default`, which is `None` by default.

    Similar to PostgreSQL's `COALESCE` function.
    """
    for value in values:
        if value is not None:
            return value
    else:
        return default


def remove_None(params: dict):
    """Remove all keys in `params` that have the value of `None`."""
    return {key: value for key, value in params.items() if value is not None}


class SpinnerContext:
    """Context of the currently running spinner."""

    def __init__(self, spinner):
        self.spinner = spinner
        self.msg = ""
        self._prev_msg = ""

    def print(self, *args, **kwargs):
        """Print inside of the spinner context.

        This must be used when inside of a spinner context to ensure that
        the line printed doesn't overwrite an already existing spinner line.
        """
        clear_len = max(len(self._prev_msg), len(self.msg)) + 4
        self.spinner.stream.write("%s\r" % (" " * clear_len))
        print(*args, file=self.spinner.stream, flush=True, **kwargs)


class Spinner:
    """Display a spinner at the terminal, if it's a TTY.

    Use as a context manager.
    """

    def __init__(self, frames="/-\|", stream=sys.stdout):
        super(Spinner, self).__init__()
        self.frames = frames
        self.stream = stream

    def __enter__(self):
        self.__context = SpinnerContext(self)
        if self.stream.isatty():
            frames = cycle(self.frames)
            stream = self.stream
            done = threading.Event()

            def run():
                # Disable cursor.
                stream.write("\033[?25l")
                stream.flush()
                try:
                    # Write out successive frames (and a backspace) every 0.1
                    # seconds until done is set.
                    while not done.wait(0.1):
                        diff = len(self.__context._prev_msg) - len(self.__context.msg)
                        if diff < 0:
                            diff = 0
                        stream.write(
                            "[%s] %s%s\r"
                            % (next(frames), self.__context.msg, " " * diff)
                        )
                        self.__context._prev_msg = self.__context.msg
                        stream.flush()
                finally:
                    # Clear line and enable cursor.
                    clear_len = (
                        max(len(self.__context._prev_msg), len(self.__context.msg)) + 4
                    )
                    stream.write("%s\r" % (" " * clear_len))
                    stream.write("\033[?25h")
                    stream.flush()

            self.__done = done
            self.__thread = threading.Thread(target=run)
            self.__thread.start()
        return self.__context

    def __exit__(self, *exc_info):
        if self.stream.isatty():
            self.__done.set()
            self.__thread.join()



"""Test for `maas.client.viscera.sshkeys`."""

import random

from .. import sshkeys

from ...testing import make_string_without_spaces, TestCase

from ..testing import bind

from testtools.matchers import Equals


def make_origin():
    return bind(sshkeys.SSHKeys, sshkeys.SSHKey)


class TestSSHKeys(TestCase):
    def test__sshkeys_create(self):
        """SSHKeys.create() returns a new SSHKey."""
        SSHKeys = make_origin().SSHKeys
        key = make_string_without_spaces()
        SSHKeys._handler.create.return_value = {"id": 1, "key": key, "keysource": ""}
        SSHKeys.create(key=key)
        SSHKeys._handler.create.assert_called_once_with(key=key)

    def test__sshkeys_read(self):
        """SSHKeys.read() returns a list of SSH keys."""
        SSHKeys = make_origin().SSHKeys
        keys = [
            {
                "id": random.randint(0, 100),
                "key": make_string_without_spaces(),
                "keysource": "",
            }
            for _ in range(3)
        ]
        SSHKeys._handler.read.return_value = keys
        ssh_keys = SSHKeys.read()
        self.assertThat(len(ssh_keys), Equals(3))


class TestSSHKey(TestCase):
    def test__sshkey_read(self):
        """SSHKeys.read() returns a single SSH key."""
        SSHKey = make_origin().SSHKey
        key_id = random.randint(0, 100)
        key_dict = {"id": key_id, "key": make_string_without_spaces(), "keysource": ""}
        SSHKey._handler.read.return_value = key_dict
        self.assertThat(SSHKey.read(id=key_id), Equals(SSHKey(key_dict)))



"""Objects for accounts."""

__all__ = ["Account"]

from . import Object, ObjectType
from ..utils.creds import Credentials


class AccountType(ObjectType):
    """Metaclass for `Account`."""

    async def create_credentials(cls) -> Credentials:
        data = await cls._handler.create_authorisation_token()
        return Credentials(
            consumer_key=data["consumer_key"],
            token_key=data["token_key"],
            token_secret=data["token_secret"],
        )

    async def delete_credentials(cls, credentials: Credentials) -> None:
        await cls._handler.delete_authorisation_token(token_key=credentials.token_key)


class Account(Object, metaclass=AccountType):
    """The current account."""



# Copyright 2016-2017 Canonical Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Encoding of MIME multipart data."""

__all__ = ["encode_multipart_data"]

from collections.abc import Iterable, Mapping
from email.generator import BytesGenerator
from email.mime.application import MIMEApplication
from email.mime.multipart import MIMEMultipart
from io import BytesIO, IOBase
from itertools import chain
import mimetypes


def get_content_type(*names):
    """Return the MIME content type for the file with the given name."""
    for name in names:
        if name is not None:
            mimetype, encoding = mimetypes.guess_type(name)
            if mimetype is not None:
                if isinstance(mimetype, bytes):
                    return mimetype.decode("ascii")
                else:
                    return mimetype
    else:
        return "application/octet-stream"


def make_bytes_payload(name, content):
    payload = MIMEApplication(content)
    payload.add_header("Content-Disposition", "form-data", name=name)
    return payload


def make_string_payload(name, content):
    payload = MIMEApplication(content.encode("utf-8"), charset="utf-8")
    payload.add_header("Content-Disposition", "form-data", name=name)
    payload.set_type("text/plain")
    return payload


def make_file_payload(name, content):
    payload = MIMEApplication(content.read())
    payload.add_header("Content-Disposition", "form-data", name=name, filename=name)
    names = name, getattr(content, "name", None)
    payload.set_type(get_content_type(*names))
    return payload


def make_payloads(name, content):
    """Constructs payload(s) for the given `name` and `content`.

    If `content` is a byte string, this calls `make_bytes_payload` to
    construct the payload, which this then yields.

    If `content` is a unicode string, this calls `make_string_payload`.

    If `content` is file-like -- it inherits from `IOBase` or `file` --
    this calls `make_file_payload`.

    If `content` is iterable, this calls `make_payloads` for each item,
    with the same name, and then re-yields each payload generated.

    If `content` is callable, this calls it with no arguments, and then
    uses the result as a context manager. This can be useful if the
    callable returns an open file, for example, because the context
    protocol means it will be closed after use.

    This raises `AssertionError` if it encounters anything else.
    """
    if content is None:
        yield make_bytes_payload(name, b"")
    elif isinstance(content, bool):
        if content:
            yield make_bytes_payload(name, b"true")
        else:
            yield make_bytes_payload(name, b"false")
    elif isinstance(content, int):
        yield make_bytes_payload(name, b"%d" % content)
    elif isinstance(content, bytes):
        yield make_bytes_payload(name, content)
    elif isinstance(content, str):
        yield make_string_payload(name, content)
    elif isinstance(content, IOBase):
        yield make_file_payload(name, content)
    elif callable(content):
        with content() as content:
            for payload in make_payloads(name, content):
                yield payload
    elif isinstance(content, Iterable):
        for part in content:
            for payload in make_payloads(name, part):
                yield payload
    else:
        raise AssertionError("%r is unrecognised: %r" % (name, content))


def build_multipart_message(data):
    message = MIMEMultipart("form-data")
    for name, content in data:
        for payload in make_payloads(name, content):
            message.attach(payload)
    return message


def encode_multipart_message(message):
    # The message must be multipart.
    assert message.is_multipart()
    # The body length cannot yet be known.
    assert "Content-Length" not in message
    # So line-endings can be fixed-up later on, component payloads must have
    # no Content-Length and their Content-Transfer-Encoding must be base64
    # (and not quoted-printable, which Django doesn't appear to understand).
    for part in message.get_payload():
        assert "Content-Length" not in part
        assert part["Content-Transfer-Encoding"] == "base64"
    # Flatten the message without headers.
    buf = BytesIO()
    generator = BytesGenerator(buf, False)  # Don't mangle "^From".
    generator._write_headers = lambda self: None  # Ignore.
    generator.flatten(message)
    # Ensure the body has CRLF-delimited lines. See
    # http://bugs.python.org/issue1349106.
    body = b"\r\n".join(buf.getvalue().splitlines())
    # Only now is it safe to set the content length.
    message.add_header("Content-Length", "%d" % len(body))
    return message.items(), body


def encode_multipart_data(data=(), files=()):
    """Create a MIME multipart payload from L{data} and L{files}.

    **Note** that this function is deprecated. Use `build_multipart_message`
    and `encode_multipart_message` instead.

    @param data: A mapping of names (ASCII strings) to data (byte string).
    @param files: A mapping of names (ASCII strings) to file objects ready to
        be read.
    @return: A 2-tuple of C{(body, headers)}, where C{body} is a a byte string
        and C{headers} is a dict of headers to add to the enclosing request in
        which this payload will travel.
    """
    if isinstance(data, Mapping):
        data = data.items()
    if isinstance(files, Mapping):
        files = files.items()
    message = build_multipart_message(chain(data, files))
    headers, body = encode_multipart_message(message)
    return body, dict(headers)



"""Testing server."""

__all__ = ["ApplicationBuilder"]

import asyncio
from collections import defaultdict
from functools import partial
import json
from operator import attrgetter
import re
from urllib.parse import urlparse

from aiohttp.multipart import CONTENT_DISPOSITION, parse_content_disposition
import aiohttp.web
from multidict import MultiDict

from . import desc


class ApplicationBuilder:
    def __init__(self, description):
        super(ApplicationBuilder, self).__init__()
        self._description = desc.Description(description)
        self._application = aiohttp.web.Application()
        (
            self._rootpath,
            self._basepath,
            self._version,
        ) = self._discover_version_and_paths()
        self._wire_up_description()
        self._actions = {}
        self._views = {}

    def route(self, method, path, handler=None):
        """Add a handler for a specific path.

        The path must start with a forward slash or be empty. The root URL of
        the server is automatically discovered from the API description and is
        added as a prefix to `path`, i.e. it's correct to specify "/accounts/"
        but incorrect to specify "/MAAS/accounts/".

        The handler is wrapped by `_wrap_handler` but should otherwise a normal
        `aiohttp` request handler.

        This method can be used as a decorator by omitting `handler`:

          @builder.route("GET", "/foo/bar")
          async def foobar(request):
              ...

        The use of `handle` should be preferred to `route` where the endpoint
        is part of MAAS's Web API. This method allows mocking of miscellaneous
        other paths when absolutely necessary.
        """
        if handler is None:
            return partial(self.route, method, path)

        if not path.startswith("/"):
            raise ValueError("Path should start with / or be empty.")

        handler = self._wrap_handler(handler)
        path = "%s/%s" % (self._rootpath, path.lstrip("/"))
        self._application.router.add_route(method, path, handler)

    def handle(self, action_name, handler=None):
        """Add a handler for a specific API action.

        The action string describes an action from the server's API
        description document. Some examples:

          auth:Machines.allocate
          anon:Version.read

        The handler is wrapped by `_wrap_handler` but should otherwise a normal
        `aiohttp` request handler.

        This method can be used as a decorator by omitting `handler`:

          @builder.handle("anon:Version.read")
          async def version(request):
              ...

        """
        if handler is None:
            return partial(self.handle, action_name)

        action = self._resolve_action(action_name)
        view_name = self._view_name(action)
        assert view_name not in self._actions
        self._actions[view_name] = action
        handler = self._wrap_handler(handler)
        if view_name in self._views:
            view = self._views[view_name]
            view.set(action, handler)
        else:
            view = self._views[view_name] = ApplicationView()
            self._application.router.add_route("*", action.path, view)
            view.set(action, handler)

    def serve(self):
        """Return an async context manager to serve the built application."""
        return ApplicationRunner(self._application, self._basepath)

    @staticmethod
    def _wrap_handler(handler):
        """Wrap `handler` in some conveniences.

        These are:

        * Setting `request.params` to a `MultiDict` instance of POSTed form
          parameters, or `None` if the body content was not a multipart form.

        * Passing `request.match_info` as keyword arguments into the handler.
          For example, if a route like "/foo/{bar}" is matched by a path
          "/foo/thing", the handler will be called with bar="thing".

        * Objects returned from `handler` that are not proper responses are
          rendered as JSON.

        """

        async def wrapper(request):
            # For convenience, read in all multipart parameters.
            assert not hasattr(request, "params")
            if request.content_type == "multipart/form-data":
                request.params = await _get_multipart_params(request)
            else:
                request.params = None
            response = await handler(request, **request.match_info)
            # For convenience, assume non-Responses are meant as JSON.
            if not isinstance(response, aiohttp.web.Response):
                response = aiohttp.web.json_response(response)
            return response

        return wrapper

    @staticmethod
    def _view_name(action):
        """Return the view name for an API action, e.g. "Version.read"."""
        return "%s.%s" % (action.resource["name"], action.name)

    def _discover_version_and_paths(self):
        """Return the root path, the API path, and the API version string.

        As an example, given an API description document containing references
        to "/MAAS/api/2.0/foo/bar", this function will return:

          ("/MAAS", "/MAAS/api/2.0", "2.0")

        """
        for resource in self._description:
            path = urlparse(resource["uri"]).path
            match = re.match("((.*)/api/([0-9.]+))/", path)
            if match is not None:
                base, root, version = match.groups()
                return root, base, version
        else:
            raise ValueError("Could not discover version or paths.")

    def _wire_up_description(self):
        """Arrange for the API description document to be served.

        This publishes only endpoints for which handlers have been registered
        using `handle`.
        """
        path = "%s/describe/" % self._basepath

        def describe(request):
            description = self._render_description(request.url.with_path(""))
            description_json = json.dumps(description, indent="  ", sort_keys=True)
            return aiohttp.web.Response(
                text=description_json, content_type="application/json"
            )

        self._application.router.add_get(path, describe)

    def _render_description(self, base):
        """Render an API description document for this application.

        This renders only endpoints for which handlers have been registered
        using `handle`.
        """
        by_resource = defaultdict(list)
        for action in self._actions.values():
            by_resource[action.resource].append(action)

        def make_resource_skeleton():
            # The "names" set gets popped later and replaced.
            return {"anon": None, "auth": None, "names": set()}

        by_resource_name = defaultdict(make_resource_skeleton)
        for resource, actions in by_resource.items():
            res_name = resource["name"]
            res_name_raw = resource["name/raw"]
            res_desc = by_resource_name[res_name]
            res_desc["names"].add(res_name_raw)
            anon_auth = "anon" if resource["is_anonymous"] else "auth"
            assert res_desc[anon_auth] is None
            res_desc[anon_auth] = {
                "actions": [
                    {
                        "doc": action.doc.title,  # Just the title.
                        "method": action.method,
                        "name": action.name,
                        "op": action.op,
                        "restful": action.is_restful,
                    }
                    for action in actions
                ],
                "doc": resource["doc"].title,
                "name": res_name_raw,
                "params": resource["params"],
                "path": resource["path"],
                "uri": str(base) + resource["path"],
            }

        for res_desc in by_resource_name.values():
            res_names = res_desc.pop("names")
            res_desc["name"] = min(res_names, key=len)

        return {
            "doc": self._description.doc.title,
            "hash": "// not calculated //",
            "resources": list(by_resource_name.values()),
        }

    def _resolve_action(self, action_name):
        """Find information on the given action.

        For example, given "anon:Version.read" it would return a `desc.Action`
        instance describing the anonymous "read" action on "VersionHandler".
        """
        match = re.match(r"^(anon|auth):(\w+[.]\w+)$", action_name)
        if match is None:
            raise ValueError(
                "Action should be (anon|auth):Resource.action, got: %s" % (action_name,)
            )
        else:
            anon_auth, resource_name = match.groups()
            resources = getattr(self._description, anon_auth)
            try:
                action = attrgetter(resource_name)(resources)
            except AttributeError:
                raise ValueError("%s not found." % resource_name)
            else:
                assert action.action_name == action_name
                return action


class ApplicationView:
    """A meta-handler that mimics the behaviour of MAAS's Web API handlers."""

    def __init__(self):
        super(ApplicationView, self).__init__()
        self.rest, self.ops = {}, {}

    @property
    def allowed_methods(self):
        allowed_methods = frozenset(self.rest)
        if len(self.ops) == 0:
            return allowed_methods
        else:
            return allowed_methods | {aiohttp.hdrs.METH_POST}

    def set(self, action, handler):
        if action.is_restful:
            self.rest[action.method] = handler
        else:
            self.ops[action.op] = handler

    async def __call__(self, request):
        if request.method == "POST":
            op = request.rel_url.query.get("op")
            if op is None:
                handler = self.rest.get(request.method)
            else:
                handler = self.ops.get(op)
        else:
            handler = self.rest.get(request.method)

        if handler is None:
            raise aiohttp.web.HTTPMethodNotAllowed(request.method, self.allowed_methods)
        else:
            return await handler(request)


class ApplicationRunner:
    """An asynchronous context manager that starts and stops an application."""

    def __init__(self, application, basepath):
        super(ApplicationRunner, self).__init__()
        self._application = application
        self._basepath = basepath

    async def __aenter__(self):
        self._loop = asyncio.get_event_loop()
        self._handler = self._application.make_handler(loop=self._loop)
        await self._application.startup()
        self._server = await self._loop.create_server(
            self._handler, host="0.0.0.0", port=0
        )
        return "http://%s:%d/%s/" % (
            *self._server.sockets[0].getsockname(),
            self._basepath.strip("/"),
        )

    async def __aexit__(self, *exc_info):
        self._server.close()
        await self._server.wait_closed()
        await self._application.shutdown()
        await self._handler.shutdown(10.0)
        await self._application.cleanup()


async def _get_multipart_params(request):
    """Extract a mapping of parts sent in a multipart request.

    :rtype: MultiDict
    """

    def get_part_name(part):
        _, params = parse_content_disposition(part.headers.get(CONTENT_DISPOSITION))
        return params.get("name")

    def get_part_data(part):
        if part.filename is None:
            return part.text()
        else:
            return part.read(decode=True)

    params = MultiDict()
    async for part in await request.multipart():
        params.add(get_part_name(part), await get_part_data(part))

    return params



# Copyright 2016-2017 Canonical Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Test multipart MIME helpers."""

from io import BytesIO
from os import urandom

from django.conf import settings
from django.core.files.uploadhandler import MemoryFileUploadHandler
from django.http.multipartparser import MultiPartParser
from django.utils.datastructures import MultiValueDict
from testtools.matchers import EndsWith, StartsWith

from ...testing import make_string, TestCase
from ..multipart import encode_multipart_data, get_content_type

# Django, sigh, needs this.
settings.configure()


ahem_django_ahem = (
    "If the mismatch appears to be because the parsed values "
    "are base64 encoded, then check you're using a >=1.4 release "
    "of Django."
)


def parse_headers_and_body_with_django(headers, body):
    """Parse `headers` and `body` with Django's :class:`MultiPartParser`.

    `MultiPartParser` is a curiously ugly and RFC non-compliant concoction.

    Amongst other things, it coerces all field names, field data, and
    filenames into Unicode strings using the "replace" error strategy, so be
    warned that your data may be silently mangled.

    It also, in 1.3.1 at least, does not recognise any transfer encodings at
    *all* because its header parsing code was broken.

    I'm also fairly sure that it'll fall over on headers than span more than
    one line.

    In short, it's a piece of code that inspires little confidence, yet we
    must work with it, hence we need to round-trip test multipart handling
    with it.
    """
    handler = MemoryFileUploadHandler()
    meta = {
        "CONTENT_TYPE": headers["Content-Type"],
        "CONTENT_LENGTH": headers["Content-Length"],
    }
    parser = MultiPartParser(
        META=meta, input_data=BytesIO(body), upload_handlers=[handler]
    )
    return parser.parse()


class TestMultiPart(TestCase):
    def test_get_content_type_guesses_type(self):
        guess = get_content_type("text.txt")
        self.assertEqual("text/plain", guess)
        self.assertIsInstance(guess, str)

    def test_encode_multipart_data_produces_bytes(self):
        data = {make_string(): make_string().encode("ascii")}
        files = {make_string(): BytesIO(make_string().encode("ascii"))}
        body, headers = encode_multipart_data(data, files)
        self.assertIsInstance(body, bytes)

    def test_encode_multipart_data_closes_with_closing_boundary_line(self):
        data = {"foo": make_string().encode("ascii")}
        files = {"bar": BytesIO(make_string().encode("ascii"))}
        body, headers = encode_multipart_data(data, files)
        self.assertThat(body, EndsWith(b"--"))

    def test_encode_multipart_data(self):
        # The encode_multipart_data() function should take a list of
        # parameters and files and encode them into a MIME
        # multipart/form-data suitable for posting to the MAAS server.
        params = {
            "op": {"value": "add", "output": "add"},
            "foo": {"value": "bar\u1234", "output": "bar\u1234"},
            "none": {"value": None, "output": ""},
            "true": {"value": True, "output": "true"},
            "false": {"value": False, "output": "false"},
            "int": {"value": 1, "output": "1"},
            "bytes": {"value": b"bytes", "output": "bytes"},
        }
        random_data = urandom(32)
        files = {"baz": BytesIO(random_data)}
        body, headers = encode_multipart_data(
            {key: value["value"] for key, value in params.items()}, files
        )
        self.assertEqual("%s" % len(body), headers["Content-Length"])
        self.assertThat(
            headers["Content-Type"], StartsWith("multipart/form-data; boundary=")
        )
        # Round-trip through Django's multipart code.
        post, files = parse_headers_and_body_with_django(headers, body)
        self.assertEqual(
            {name: [value["output"]] for name, value in params.items()},
            post,
            ahem_django_ahem,
        )
        self.assertSetEqual({"baz"}, set(files))
        self.assertEqual(random_data, files["baz"].read(), ahem_django_ahem)

    def test_encode_multipart_data_multiple_params(self):
        # Sequences of parameters and files passed to
        # encode_multipart_data() permit use of the same name for
        # multiple parameters and/or files. See `make_payloads` to
        # understand how it processes different types of parameter
        # values.
        params_in = [("one", "ABC"), ("one", "XYZ"), ("two", ["DEF", "UVW"])]
        files = [
            BytesIO(b"f1"),
            self.makeFile(contents=b"f2").open("rb"),
            self.makeFile(contents=b"f3").open("rb"),
        ]
        for fd in files:
            self.addCleanup(fd.close)
        files_in = [
            ("f-one", files[0]),
            ("f-two", files[1]),
            ("f-three", lambda: files[2]),
        ]
        body, headers = encode_multipart_data(params_in, files_in)
        self.assertEqual("%s" % len(body), headers["Content-Length"])
        self.assertThat(
            headers["Content-Type"], StartsWith("multipart/form-data; boundary=")
        )
        # Round-trip through Django's multipart code.
        params_out, files_out = parse_headers_and_body_with_django(headers, body)
        params_out_expected = MultiValueDict()
        params_out_expected.appendlist("one", "ABC")
        params_out_expected.appendlist("one", "XYZ")
        params_out_expected.appendlist("two", "DEF")
        params_out_expected.appendlist("two", "UVW")
        self.assertEqual(params_out_expected, params_out, ahem_django_ahem)
        files_expected = {"f-one": b"f1", "f-two": b"f2", "f-three": b"f3"}
        files_observed = {name: buf.read() for name, buf in files_out.items()}
        self.assertEqual(files_expected, files_observed, ahem_django_ahem)

    def test_encode_multipart_data_list_params(self):
        params_in = [("one", ["ABC", "XYZ"]), ("one", "UVW")]
        body, headers = encode_multipart_data(params_in, [])
        params_out, files_out = parse_headers_and_body_with_django(headers, body)
        self.assertEqual({"one": ["ABC", "XYZ", "UVW"]}, params_out)
        self.assertSetEqual(set(), set(files_out))



"""Objects for MAASs."""

__all__ = ["MAAS"]

import enum
import re
import typing

from . import Object, ObjectType
from ..bones import CallError


def _django_boolean(boolean):
    """Render a string suitable for use in a Django form.

    Django's `BooleanField` understands "true", "false", "1", "0", "t", and
    "f" as valid form values, but the widget `CheckboxInput` that is used by
    `BooleanField` by default gets a shot at converting the form input first.
    It inexplicably has different rules. See in `value_from_datadict`:

      values = {'true': True, 'false': False}
      if isinstance(value, six.string_types):
          value = values.get(value.lower(), value)
      return bool(value)

    Sigh.
    """
    return "true" if boolean else "false"


class DescriptiveEnum(enum.Enum):
    """An enum where each member is a `(parameter, description)` tuple."""

    @property
    def parameter(self):
        return self.value[0]

    @property
    def description(self):
        return self.value[1]

    @classmethod
    def lookup(cls, parameter):
        for member in cls:
            if member.parameter == parameter:
                return member
        else:
            raise KeyError("%s value %r not recognised." % (cls.__name__, parameter))


class MAASType(ObjectType):
    """Metaclass for `MAAS`."""

    async def get_name(cls) -> str:
        """The name of the MAAS instance."""
        return await cls.get_config("maas_name")

    async def set_name(cls, name: str):
        """See `get_name`."""
        return await cls.set_config("maas_name", name)

    async def get_main_archive(cls) -> str:
        """Main archive URL.

        Archive used by nodes to retrieve packages for Intel architectures,
        e.g. http://archive.ubuntu.com/ubuntu.
        """
        return await cls.get_config("main_archive")

    async def set_main_archive(cls, url: str):
        """See `get_main_archive`."""
        await cls.set_config("main_archive", url)

    async def get_ports_archive(cls) -> str:
        """Ports archive.

        Archive used by nodes to retrieve packages for non-Intel
        architectures. E.g. http://ports.ubuntu.com/ubuntu-ports.
        """
        return await cls.get_config("ports_archive")

    async def set_ports_archive(cls, series: str):
        """See `get_ports_archive`."""
        await cls.set_config("ports_archive", series)

    async def get_default_os(cls) -> str:
        """Default OS used for deployment."""
        return await cls.get_config("default_osystem")

    async def set_default_os(cls, series: str):
        """See `get_default_os`."""
        await cls.set_config("default_osystem", series)

    async def get_default_distro_series(cls) -> str:
        """Default OS release used for deployment."""
        return await cls.get_config("default_distro_series")

    async def set_default_distro_series(cls, series: str):
        """See `get_default_distro_series`."""
        await cls.set_config("default_distro_series", series)

    async def get_commissioning_distro_series(cls) -> str:
        """Default Ubuntu release used for commissioning."""
        return await cls.get_config("commissioning_distro_series")

    async def set_commissioning_distro_series(cls, series: str):
        """See `get_commissioning_distro_series`."""
        await cls.set_config("commissioning_distro_series", series)

    async def get_http_proxy(cls) -> typing.Optional[str]:
        """Proxy for APT and HTTP/HTTPS.

        This will be passed onto provisioned nodes to use as a proxy for APT
        traffic. MAAS also uses the proxy for downloading boot images. If no
        URL is provided, the built-in MAAS proxy will be used.
        """
        data = await cls.get_config("http_proxy")
        return None if data is None or data == "" else data

    async def set_http_proxy(cls, url: typing.Optional[str]):
        """See `get_http_proxy`."""
        await cls.set_config("http_proxy", "" if url is None else url)

    async def get_enable_http_proxy(cls) -> bool:
        """Enable the use of an APT and HTTP/HTTPS proxy.

        Provision nodes to use the built-in HTTP proxy (or user specified
        proxy) for APT. MAAS also uses the proxy for downloading boot images.
        """
        return await cls.get_config("enable_http_proxy")

    async def set_enable_http_proxy(cls, enabled: bool):
        """See `get_enable_http_proxy`."""
        await cls.set_config("enable_http_proxy", _django_boolean(enabled))

    async def get_curtin_verbose(cls) -> bool:
        """Should `curtin` log with high verbosity?

        Run the fast-path installer with higher verbosity. This provides more
        detail in the installation logs.
        """
        return await cls.get_config("curtin_verbose")

    async def set_curtin_verbose(cls, verbose: bool):
        """See `get_curtin_verbose`."""
        await cls.set_config("curtin_verbose", _django_boolean(verbose))

    async def get_kernel_options(cls) -> typing.Optional[str]:
        """Kernel options.

        Boot parameters to pass to the kernel by default.
        """
        data = await cls.get_config("kernel_opts")
        return None if data is None or data == "" else data

    async def set_kernel_options(cls, options: typing.Optional[str]):
        """See `get_kernel_options`."""
        await cls.set_config("kernel_opts", "" if options is None else options)

    async def get_upstream_dns(cls) -> list:
        """Upstream DNS server addresses.

        Upstream DNS servers used to resolve domains not managed by this MAAS
        (space-separated IP addresses). Only used when MAAS is running its own
        DNS server. This value is used as the value of 'forwarders' in the DNS
        server config.
        """
        data = await cls.get_config("upstream_dns")
        return [] if data is None else re.split(r"[,\s]+", data)

    async def set_upstream_dns(cls, addresses: typing.Optional[typing.Sequence[str]]):
        """See `get_upstream_dns`."""
        await cls.set_config(
            "upstream_dns", ("" if addresses is None else ",".join(addresses))
        )

    class DNSSEC(DescriptiveEnum):
        """DNSSEC validation settings.

        See `get_dnssec_validation` and `set_dnssec_validation`.
        """

        AUTO = "auto", "Automatic (use default root key)"
        YES = "yes", "Yes (manually configured root key)"
        NO = "no", "No (disable DNSSEC)"

    async def get_dnssec_validation(cls) -> DNSSEC:
        """Enable DNSSEC validation of upstream zones.

        Only used when MAAS is running its own DNS server. This value is used
        as the value of 'dnssec_validation' in the DNS server config.
        """
        data = await cls.get_config("dnssec_validation")
        return cls.DNSSEC.lookup(data)

    async def set_dnssec_validation(cls, validation: DNSSEC):
        """See `get_dnssec_validation`."""
        await cls.set_config("dnssec_validation", validation.parameter)

    async def get_default_dns_ttl(cls) -> int:
        """Default Time-To-Live for DNS records.

        If no TTL value is specified at a more specific point this is how long
        DNS responses are valid, in seconds.
        """
        return int(await cls.get_config("default_dns_ttl"))

    async def set_default_dns_ttl(cls, ttl: int):
        """See `get_default_dns_ttl`."""
        await cls.set_config("default_dns_ttl", str(ttl))

    async def get_enable_disk_erasing_on_release(cls) -> bool:
        """Should nodes' disks be erased prior to releasing."""
        return await cls.get_config("enable_disk_erasing_on_release")

    async def set_enable_disk_erasing_on_release(cls, erase: bool):
        """Should nodes' disks be erased prior to releasing."""
        await cls.set_config("enable_disk_erasing_on_release", _django_boolean(erase))

    async def get_windows_kms_host(cls) -> typing.Optional[str]:
        """Windows KMS activation host.

        FQDN or IP address of the host that provides the KMS Windows
        activation service. (Only needed for Windows deployments using KMS
        activation.)
        """
        data = await cls.get_config("windows_kms_host")
        return None if data is None or data == "" else data

    async def set_windows_kms_host(cls, host: typing.Optional[str]):
        """See `get_windows_kms_host`."""
        await cls.set_config("windows_kms_host", "" if host is None else host)

    async def get_boot_images_auto_import(cls) -> bool:
        """Automatically import/refresh the boot images every 60 minutes."""
        return await cls.get_config("boot_images_auto_import")

    async def set_boot_images_auto_import(cls, auto: bool):
        """See `get_boot_images_auto_import`."""
        await cls.set_config("boot_images_auto_import", _django_boolean(auto))

    async def get_ntp_server(cls) -> str:
        """Address of NTP server.

        NTP server address passed to nodes via a DHCP response. e.g.
        ntp.ubuntu.com.
        """
        return await cls.get_config("ntp_server")

    async def set_ntp_server(cls, server: str):
        """See `get_ntp_server`."""
        await cls.set_config("ntp_server", server)

    class StorageLayout(DescriptiveEnum):

        FLAT = "flat", "Flat layout"
        LVM = "lvm", "LVM layout"
        BCACHE = "bcache", "Bcache layout"

    async def get_default_storage_layout(cls) -> StorageLayout:
        """Default storage layout.

        Storage layout that is applied to a node when it is deployed.
        """
        data = await cls.get_config("default_storage_layout")
        return cls.StorageLayout.lookup(data)

    async def set_default_storage_layout(cls, series: StorageLayout):
        """See `get_default_storage_layout`."""
        await cls.set_config("default_storage_layout", series.parameter)

    async def get_default_min_hwe_kernel(cls) -> typing.Optional[str]:
        """Default minimum kernel version.

        The minimum kernel version used on new and commissioned nodes.
        """
        data = await cls.get_config("default_min_hwe_kernel")
        return None if data is None or data == "" else data

    async def set_default_min_hwe_kernel(cls, version: typing.Optional[str]):
        """See `get_default_min_hwe_kernel`."""
        await cls.set_config(
            "default_min_hwe_kernel", "" if version is None else version
        )

    async def get_enable_third_party_drivers(cls) -> bool:
        """Enable the installation of proprietary drivers, e.g. HPVSA."""
        return await cls.get_config("enable_third_party_drivers")

    async def set_enable_third_party_drivers(cls, enabled: bool):
        """See `get_enable_third_party_drivers`."""
        await cls.set_config("enable_third_party_drivers", _django_boolean(enabled))

    async def get_config(cls, name: str):
        """Get a configuration value from MAAS.

        Consult your MAAS server for recognised settings. Alternatively, use
        the pre-canned functions also defined on this object.
        """
        return await cls._handler.get_config(name=[name])

    async def set_config(cls, name: str, value):
        """Set a configuration value in MAAS.

        Consult your MAAS server for recognised settings. Alternatively, use
        the pre-canned functions also defined on this object.
        """
        return await cls._handler.set_config(name=[name], value=[value])

    async def _roundtrip(cls):
        """Testing helper: gets each value and sets it again."""
        getters = {
            name[4:]: getattr(cls, name)
            for name in dir(cls)
            if name.startswith("get_") and name != "get_config"
        }
        setters = {
            name[4:]: getattr(cls, name)
            for name in dir(cls)
            if name.startswith("set_") and name != "set_config"
        }

        for name, getter in getters.items():
            print(">>>", name)
            value = await getter()
            print(" ->", repr(value))
            print("   ", type(value))
            setter = setters[name]
            try:
                await setter(value)
            except CallError as error:
                print(error)
                print(error.content.decode("utf-8", "replace"))
            else:
                value2 = await getter()
                if value2 != value:
                    print("!!! Round-trip failed:", repr(value), "-->", repr(value2))


class MAAS(Object, metaclass=MAASType):
    """The current MAAS."""
